'''
evaluate gene statistics of the state-specific GRNs
'''

import numpy as np
import pandas as pd
import networkx as nx
from scipy.stats import iqr

def signaling_score(adata):
    '''Compute the signaling scores of all genes in each cell states

    Scores are defined based on number of incoming/outgoing edges and weighted sum of incoming/outgoing edges
    Results are saved in  adata.uns['signaling_scores']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix

    Returns
    -------
    None

    '''
    types = sorted(list(set(list(adata.obs['clusters']))))
    n = len(list(adata.var_names))

    jac_df = adata.uns['average_jac']
    sign_score = {}

    for t in types:
        M = jac_df[t][0][0:n, n:].copy()
        rec, sen = np.sum(np.abs(M), axis=1), np.sum(np.abs(M), axis=0)

        rec_edges, sen_edges = np.zeros(M[0].size), np.zeros(M[0].size)
        for i in range(rec_edges.size):
            rec_edges[i] = np.count_nonzero(np.abs(M[i]))
            sen_edges[i] = np.count_nonzero(np.abs(M[:,i]))

        sign_score[t] = {'incoming': rec, 'outgoing': sen, 'incoming_edges': rec_edges, 'outgoing_edges': sen_edges}
    adata.uns['signaling_scores'] = sign_score

def compute_metrics(dat):
    '''
    Compute statistical metrics including standard deviation, range, and interquartile range of a 1D vector

    Parameters
    ----------
    dat: `~numpy.ndarray`
        1D array of measurements

    Returns
    -------
    metrics_df: `~pandas.dataFrame`
        dataframe of metrics

    '''
    SD, rng, int_qrt = np.zeros(dat[0].size), np.zeros(dat[0].size), np.zeros(dat[0].size)

    SD = np.std(dat, axis=0)

    for i in range(rng.size):
        rng[i] = np.amax(dat[:, i]) - np.amin(dat[:, i])

    int_qrt = iqr(dat, axis=0)

    d = {'SD':SD, 'range':rng, 'inter_range':int_qrt}
    metrics_df = pd.DataFrame(d)
    return metrics_df


def GRN_cluster_variation(adata):
    '''Compute state-specific GRN statistics including gene centrality, incoming, outgoing and total signaling strength
    Results are stored in adata.uns['cluster_variation'] and adata.uns['cluster_average']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix

    Returns
    -------
    None

    '''
    types = sorted(list(set(list(adata.obs['clusters']))))
    genes = list(adata.var_names)

    dat_BC, dat_in, dat_out, dat_tot = np.zeros((len(types), len(genes))), np.zeros((len(types), len(genes))), np.zeros((len(types), len(genes))), np.zeros((len(types), len(genes)))

    for i in range(len(types)):
        dat_BC[i] = np.asarray(adata.uns['GRN_statistics'][0][types[i]].mean(axis=1))
        dat_in[i] = np.asarray(adata.uns['GRN_statistics'][1][types[i]].mean(axis=1))
        dat_out[i] = np.asarray(adata.uns['GRN_statistics'][2][types[i]].mean(axis=1))
        dat_tot[i] = np.asarray(adata.uns['GRN_statistics'][3][types[i]].mean(axis=1))

    gene_average = { 'centrality': dat_BC.mean(axis=0), 'incoming': dat_in.mean(axis=0), 'outgoing': dat_out.mean(axis=0), 'signaling': dat_tot.mean(axis=0) }
    adata.uns['cluster_average'] = gene_average

    gene_variation = {}
    gene_variation['centrality'] = compute_metrics(dat_BC)
    gene_variation['incoming'] = compute_metrics(dat_in)
    gene_variation['outgoing'] = compute_metrics(dat_out)
    gene_variation['signaling'] = compute_metrics(dat_tot)
    adata.uns['cluster_variation'] = gene_variation

def grn_statistics(adata,
                   weight_quantile=0.5,
                   k=None,
                   normalized=True,
                   weight=None,
                   endpoints=False,
                   seed=None
                   ):
    '''
    Computes various statistics on the cell state specific GRNs. The statistics are added to adata.uns['GRN_statistics']

    For a more detailed discussion of several of the parameters, please see the betweenness_centrality from Networkx
    (https://networkx.org/documentation/networkx-1.10/reference/generated/networkx.algorithms.centrality.betweenness_centrality.html)
    Results are stored in adata.uns['GRN_statistics'], adata.uns['cluster_variation'] and adata.uns['cluster_average']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    weight_quantile: `float` (default: 0.5)
        cutoff for weak gene-gene interactions between 0 and 1
    k: `int` or `None` (default=`None`)
        number of nodes considered to compute betweenness centrality. k=None implies that all edges are used
    normalized: `Bool` (default: True)
        if True, betweenness centrality values are normalized
    weight: `str` (default: `None`)
        If None, all edge weights are considered equal
    endpoints: `Bool` (default: False)
        If True, include the endpoints in the shortest path counts during betweenness centrality calculation
    seed: `int` (default: None)
        seed for betweenness centrality calculation

    Returns
    -------
    None

    '''
    assert 'jacobian_lists' in adata.uns.keys(), "Please run 'tl.estimate_jacobian' before calling 'grn_statistics'"

    genes = list(adata.var_names)
    n = len(genes)

    jacobian_lists = adata.uns['jacobian_lists']
    types = jacobian_lists.keys()

    betwenness_cent, incoming, outgoing, total_sign = {}, {}, {}, {}

    # repeat analysis for all clusters
    for t in types:
        df_bc, df_inc = pd.DataFrame( index=genes ), pd.DataFrame( index=genes )
        df_out, df_tot = pd.DataFrame( index=genes ), pd.DataFrame( index=genes )

        jacs = jacobian_lists[t][0]
        for j in range(len(jacs)):
            # extract 'interacting' quadrant of the Jacobian
            A = jacs[j][0:n, n:].copy().T

            rec, sen = np.sum(np.abs(A), axis=0), np.sum(np.abs(A), axis=1)
            tot_sign = rec + sen

            q_pos = np.quantile(A[A > 0], weight_quantile)
            q_neg = np.quantile(A[A < 0], 1 - weight_quantile)
            A[(A > q_neg) & (A < q_pos)] = 0
            G = nx.from_numpy_matrix(A, create_using=nx.DiGraph)
            centrality = nx.betweenness_centrality(G, k=k, normalized=normalized, weight=weight, endpoints=endpoints, seed=seed)

            df_bc[j] = centrality.values()
            df_inc[j] = rec
            df_out[j] = sen
            df_tot[j] = tot_sign

        betwenness_cent[t] = df_bc
        incoming[t] = df_inc
        outgoing[t] = df_out
        total_sign[t] = df_tot

    adata.uns['GRN_statistics'] = [betwenness_cent, incoming, outgoing, total_sign]

    GRN_cluster_variation(adata)

    # compute signaling scores of individual genes in each cell state
    signaling_score(adata)


'''
test the sensitivity to the regression method and parameters
'''

import numpy as np

from .aux_functions import parameter_regression

def coeff_dist(G,
               x = np.logspace(-5, 0, num=100)
               ):
    '''Construct the coefficient distribution for the gene-gene interaction matrix G

    Parameters
    ----------
    G: `~numpy.ndarray`
        interaction matrix
    x: `~numpy.ndarray` (default: numpy.logspace(-5, 0, num=100))
        vector of thresholds to group matrix coefficients

    Returns
    -------
    y: `~numpy.ndarray`
        The distribution of coefficients

    '''
    pars = np.abs(np.ndarray.flatten(G))
    coeffs = pars/np.amax(pars)
    y = np.zeros(x.size)
    for i in range(x.size):
        y[i] = coeffs[coeffs > x[i]].size
    y = y/pars.size
    return y


def regr_method_sens(adata,
                     alpha_ridge = np.array([0.01, 0.1, 1, 10, 100]),
                     alpha_lasso = np.array([0.0001, 0.001, 0.01, 0.1, 1])
                     ):
    '''Compare methods for gene-gene interaction parameter regression
    Results are stored in adata.uns['method_sens'] and adata.uns['sens_coeff']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    alpha_ridge: `~numpy.ndarray` (default: numpy.array([0.01, 0.1, 1, 10, 100]))
        array of shrinkage coefficients to test for Ridge regression
    alpha_lasso: `~numpy.ndarray` (default: numpy.array([0.0001, 0.001, 0.01, 0.1, 1]))
        array of shrinkage coefficients to test for Lasso regression

    Returns
    -------
    None

    '''
    types = list(set(list(adata.obs['clusters'])))

    method_sens = {}

    x = np.logspace(-5, 0, num=100)
    y_reg, y_ridge, y_lasso = np.zeros(x.size), np.zeros((alpha_ridge.size, x.size)), np.zeros((alpha_lasso.size, x.size))

    for i in range(len(types)):
        print('Running method sensitivity on the ' + types[i] + ' cluster...')

        ridge_jac, lasso_jac = [], []

        sel_adata = adata[adata.obs['clusters'] == types[i]]
        U = sel_adata.layers['unspliced'].toarray()
        S = sel_adata.layers['spliced'].toarray()

        # linear regression
        B_lin, C, G = parameter_regression(U, S, method='Linear')
        y_reg = coeff_dist(B_lin, x=x)

        # Ridge regression
        for j in range(alpha_ridge.size):
            B_ridge, C, G = parameter_regression(U, S, alpha=alpha_ridge[j])
            ridge_jac.append(B_ridge)
            y_ridge[j] = coeff_dist(B_ridge)

        # Lasso regression
        for j in range(alpha_lasso.size):
            B_lasso, C, G = parameter_regression(U, S, method='Lasso', alpha=alpha_lasso[j])
            lasso_jac.append(B_lasso)
            y_lasso[j] = coeff_dist(B_lasso)

        method_sens[types[i]] = [B_lin, ridge_jac, lasso_jac]
    adata.uns['method_sens'] = [alpha_ridge, alpha_lasso, method_sens]
    adata.uns['sens_coeff'] = [x, y_reg, y_ridge, y_lasso]


'''
functions to plot gene variation across cell states
'''
import numpy as np
import matplotlib.pyplot as plt

def set_plot_name(method,
                  measure
                  ):
    '''Set the plot name based on method and measure used

    Parameters
    ----------
    method: `str`
        method used
    measure: `str`
        measure used

    Returns
    -------
    measure_name: `str`
        string with name of measure
    method_name: `str`
        string with name of method

    '''
    if measure=='centrality':
        measure_name = 'Betweenness Centrality'
    elif measure=='incoming':
        measure_name = 'Incoming Signaling'
    elif measure=='outgoing':
        measure_name = 'Outgoing signaling'
    else:
        measure_name = 'Total Signaling'

    if method=='SD':
        method_name = 'Standard Deviation'
    elif method=='range':
        method_name = 'Range'
    elif method=='inter_range':
        method_name='Interquartile Range'

    return measure_name, method_name


def gene_variation(adata,
                   n_genes='all',
                   method='SD',
                   measure='centrality',
                   bar_color='paleturquoise',
                   alpha=1,
                   edge_color='mediumturquoise',
                   edge_width=1,
                   gene_label_rot=90,
                   fontsize=10,
                   showfig=None,
                   savefig=None,
                   format='pdf',
                   figsize=(6, 3)
                   ):
    '''Bar plot of gene role variation across cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    n_genes: `str` or `int` (default n_genes='all')
        number of genes to consider. If an integer (n) is provided, the top n genes are selected.
        Otherwise, all genes are used if n_genes='all'
    method: `float` (default: "SD")
        method to estimate gene role variation across cell states, choose between standard deviation ('SD'),
        range ('range'), and interquartile range ('inter_range')
    measure: `float` (default: 'centrality')
        measure to estimate gene role variation, choose between 'centrality', 'incoming', 'outgoing', 'signaling'
    bar_color: `str` (default: 'paleturquoise')
        color for bar plot. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    alpha: `float` (default: 1)
        shading of bar plot between 0 and 1
    edge_color: `str` (default: 'paleturquoise')
        edge color for bar plot. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    edge_width: `float` (default: 1)
        edge width for bar plot
    gene_label_rot: `int` (default: 90)
        rotation of labels on x-axis
    fontsize: `int` (default: 10)
        fontsize of figure
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (6,3))
        size of figure

    Returns
    -------
    None

    '''
    assert 'GRN_statistics' in adata.uns.keys(), "Please run 'grn_statistics' before calling gene_variation()"
    assert measure in ['centrality', 'incoming', 'outgoing', 'signaling'], \
        "Please choose method from the list ['centrality', 'incoming', 'outgoing', 'signaling']"

    measure_name, method_name = set_plot_name(method, measure)

    genes = list(adata.var_names)
    y = adata.uns['cluster_variation'][measure][method]
    ind = np.flip(np.argsort(y))

    if n_genes=='all':
        pass
    else:
        ind = ind[0:n_genes]

    plt.figure(figsize=figsize)

    top = 15

    ax1 = plt.subplot(121)
    ax2 = plt.subplot(122, sharey=ax1)

    bars = np.asarray(y[ind])

    ax1.bar(np.arange(1, top+1, 1), bars[0:top], width=1, align='center', color=bar_color, alpha=alpha,
            edgecolor=edge_color, lw=edge_width)
    ax2.bar(np.arange(bars.size-top+1, bars.size+1, 1), bars[bars.size-top:], width=1, align='center',
            color=bar_color, alpha=alpha, edgecolor=edge_color, lw=edge_width,
            label=measure_name + '\n' + method_name + '\nacross Cell Types')

    ax1.set_xticks(np.arange(1, top + 1, 1))
    ax2.set_xticks(np.arange(bars.size-top+1, bars.size+1, 1))
    ax1.set_xticklabels([ genes[i] for i in ind[0:top] ], rotation=gene_label_rot, fontsize=fontsize)
    ax2.set_xticklabels([ genes[i] for i in ind[bars.size-top:] ], rotation=gene_label_rot, fontsize=fontsize)
    ax1.set_xlim(0, top+1)
    ax2.set_xlim(bars.size-top, bars.size+1)

    ax1.spines['right'].set_visible(False)
    ax2.spines['left'].set_visible(False)
    ax2.tick_params(left=False, labelleft=False)

    d = .015
    kwargs = dict(transform=ax1.transAxes, color='k', clip_on=False)
    ax1.plot((1 - d, 1 + d), (-d, +d), **kwargs, lw=1)  # top-left diagonal
    ax1.plot((1 - d, 1 + d), (1 - d, 1 + d), **kwargs, lw=1)  # top-right diagonal
    kwargs.update(transform=ax2.transAxes)  # switch to the bottom axes
    ax2.plot((-d, +d), (-d, +d), **kwargs, lw=1)  # bottom-left diagonal
    ax2.plot((-d, +d), (1 - d, 1 + d), **kwargs, lw=1)

    plt.legend(loc='upper right', fontsize=fontsize)
    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def gene_var_detail(adata,
                    n_genes=5,
                    select='top',
                    method='SD',
                    measure='centrality',
                    loc='best',
                    fontsize=10,
                    legend=True,
                    legend_font=10,
                    gene_label_rot=45,
                    showfig=None,
                    savefig=None,
                    format='pdf',
                    figsize=(5, 4)
                    ):
    '''Plot the detailed variation in gene signaling role for the top genes in the dataset

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    n_genes: `int` (default:5)
        number of top genes
    select: `str` (default: 'top')
        choose to select genes with larger variation between cell states (select='top')
        or small variation (select='bottom')
    method: `float` (default: "SD")
        method to estimate gene role variation across cell states, choose between standard deviation ('SD'),
        range ('range'), and interquartile range ('inter_range')
    measure: `float` (default: 'centrality')
        measure to estimate gene role variation, choose between 'centrality', 'incoming', 'outgoing', 'signaling'
    legend_loc: `str` (default: 'best')
        legend location.  Details on legend location can be found at:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
    fontsize: `int` (default: 10)
        fontsize of figure
    legend: `Bool` (default: True)
        if True, include legend
    legend_font: `int` (default: 10)
        font of legend
    gene_label_rot: `int` (default: 45)
        rotation of labels on x-axis
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (5,4))
        size of figure

    Returns
    -------
    None

    '''
    assert 'GRN_statistics' in adata.uns.keys(), "Please run 'grn_statistics' before calling gene_variation()"
    assert measure in ['centrality', 'incoming', 'outgoing','signaling'], \
        "Please choose method from the list ['centrality', 'incoming', 'outgoing', 'signaling']"

    genes = list(adata.var_names)
    types = sorted(list(set(list(adata.obs['clusters']))))
    colors = list(plt.cm.Set2.colors)[0:len(types)]

    measure_name, method_name = set_plot_name(method, measure)

    y = adata.uns['cluster_variation'][measure][method]
    ind = np.asarray(np.flip(np.argsort(y)))

    data = np.zeros((len(types), len(genes)))
    for i in range(len(types)):
        data[i] = np.asarray(adata.uns['GRN_statistics'][3][types[i]].mean(axis=1))

    x = np.arange(1, n_genes+1, 1)
    x_add = np.linspace(-0.25, 0.25, num=len(types))
    dx = x_add[1]-x_add[0]

    plt.figure(figsize=figsize)

    for i in range(len(types)):
        if select=='top':
            y = data[i, ind[0:n_genes]]
            label_names=[genes[i] for i in ind[0:n_genes]]
        elif select=='bottom':
            y = data[i, ind[ind.size-n_genes:]]
            label_names = [genes[i] for i in ind[ind.size-n_genes:]]
        plt.bar(x + x_add[i], y, color=colors[i], align='center', width=dx, label=types[i])

    plt.xticks(np.arange(1, x.size + 1, 1), label_names, fontsize=fontsize, rotation=gene_label_rot)
    plt.ylabel(measure_name, fontsize=fontsize)
    if legend:
        plt.legend(loc=loc, fontsize=legend_font)
    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def gene_var_scatter(adata,
                     method='SD',
                     measure='centrality',
                     top_genes=5,
                     fontsize=10,
                     color='b',
                     showfig=None,
                     savefig=None,
                     format='pdf',
                     figsize=(5, 4)
                     ):
    '''Scatter plot of gene role variation across cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    method: `float` (default: "SD")
        method to estimate gene role variation across cell states, choose between standard deviation ('SD'),
        range ('range'), and interquartile range ('inter_range')
    measure: `float` (default: 'centrality')
        measure to estimate gene role variation, choose between 'centrality', 'incoming', 'outgoing', 'signaling'
    top_genes: `int` (default: 5)
        top genes to annotate
    fontsize: `int` (default: 10)
        fontsize of figure
    color: `str` (default: 'b')
        color of scatter plot. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (5,4))
        size of figure

    Returns
    -------
    None

    '''
    genes = list(adata.var_names)

    measure_name, method_name = set_plot_name(method, measure)

    x, y = adata.uns['cluster_average'][measure], np.asarray(adata.uns['cluster_variation'][measure][method])

    sort_ind = np.argsort(y)
    x_top, y_top, lab = x[sort_ind[x.size - top_genes:]], y[sort_ind[x.size - top_genes:]], []

    for i in sort_ind[x.size - top_genes:]:
        lab.append(genes[i])

    plt.figure(figsize=figsize)
    ax = plt.subplot(111)

    plt.scatter(x, y, color=color)
    for i in range(x_top.size):
        ax.annotate(lab[i], (x_top[i], y_top[i]), fontsize=fontsize)
    plt.xlabel('Average across\nStates', fontsize=fontsize)
    plt.ylabel(method_name + ' across States', fontsize=fontsize)
    plt.title(measure_name)
    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def compare_standout_genes(adata,
                           cluster_list=None,
                           top_genes=5,
                           criterium='centrality',
                           panel_height=1.5,
                           panel_length=5,
                           ylabel=False,
                           showfig=None,
                           savefig=None,
                           format='pdf'
                           ):
    '''
    Boxplot to compare the standout genes with state-specific roles across many cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster_list: `str`
        list of cell states to compare
    top_genes: `str` (default: 5)
        number of top genes to consider
    criterium: `str` (default='centrality')
        measure to use to evaluate gene role, choose between 'centrality', 'incoming', 'outgoing', 'signaling'
    panel_height: `float` (default: 1.5)
        height of each panel (in inches)
    panel_length: `float` (default: 5)
        length of each panel (in inches)
    ylabel: `Bool` (default: False)
        if True, print label of y-axis
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format

    Returns
    -------
    None

    '''
    types = sorted(list(set(list(adata.obs['clusters']))))
    colors = list(plt.cm.Set3.colors)[0:len(types)]

    if cluster_list==None:
        cluster_list = types

    # set colors for plotting
    sel_colors = []
    for i in range(len(cluster_list)):
        sel_colors.append( colors[types.index(cluster_list[i])] )

    if criterium=='centrality':
        score = 0
        y_lab = 'Betweenness\nCentrality'
    elif criterium=='incoming':
        score = 1
        y_lab = 'Incoming\nSignaling'
    elif criterium=='outgoing':
        score = 2
        y_lab = 'Outgoing\nSignaling'
    elif criterium=='total':
        score = 3
        y_lab = 'Incoming+\nOutgoing'

    genes = list(adata.var_names)
    mean, fc = np.zeros((len(cluster_list), len(genes))), np.zeros((len(cluster_list), len(genes)))

    for i in range(len(cluster_list)):
        bt_df = adata.uns['GRN_statistics'][score][cluster_list[i]]
        mean[i] = bt_df.mean(axis=1)
    avg = np.mean(mean, axis=0)

    for i in range(len(cluster_list)):
        fc[i] = mean[i]/avg
    fc = np.nan_to_num(fc)
    fc = np.transpose(fc)

    # find 5 most differentiated genes based on betweenness
    j = 0
    fig = plt.figure(figsize=(panel_length,top_genes*panel_height))

    while j<top_genes:
        [a,b] = np.unravel_index(fc.argmax(), fc.shape)


        btn = []
        for i in range(len(cluster_list)):
            bt_df = adata.uns['GRN_statistics'][score][cluster_list[i]]
            btn.append(np.asarray(bt_df.loc[genes[a]]))

        ax = plt.subplot2grid((top_genes, 1), (j, 0), rowspan=1, colspan=1)
        bpt = plt.boxplot(btn, patch_artist=True)
        for patch, color in zip(bpt['boxes'], sel_colors):
            patch.set_facecolor(color)
        if ylabel:
            plt.ylabel(y_lab)

        plt.xticks([])
        plt.title(genes[a])

        j = j + 1
        fc = np.delete(fc, a, axis=0)
        genes.remove( genes[a] )

    plt.xticks(np.arange(1, len(cluster_list) + 1, 1), cluster_list)
    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)





'''
functions to quantitatively compare GRNs across cell states
'''
import numpy as np
import matplotlib.pyplot as plt

def plot_grn_comparison(adata,
                        score='AUPRC',
                        edges='all',
                        cmap='Reds',
                        title=False,
                        fontsize=10,
                        showfig=None,
                        savefig=None,
                        format='pdf',
                        figsize=(5,4)
                        ):
    '''Plot a heatmap of pairwise similarities between GRNs of different cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    score: `str` (default: 'AUPRC')
        metric to use, choose between 'AUROC' and 'AUPRC'
    edges: `str` (default: 'all')
        which edges to consider for similarity, choose between 'all', 'positive', 'negative'
    cmap: `pyplot colormap` (default='Reds')
        colormap to use. A list of accepted colormaps can be found at:
        https://matplotlib.org/stable/tutorials/colors/colormaps.html
    title: `str` or `Bool` (default='False')
        plot title, must be provided as a string
    fontsize: `int` (default: 10)
        fontsize for figure
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (5,4))
        size of figure

    Returns
    -------
    None

    '''
    assert score in ['AUROC', 'AUPRC'], 'Choose between score="AUROC" and score="AUPRC"'
    assert edges in ['all', 'positive', 'negative'], 'Choose between edges="all", "positive" or "negative"'

    types = sorted(list(set(list(adata.obs['clusters']))))

    mat = adata.uns['comparison_scores'][score + '_' + edges]

    plt.figure(figsize=figsize)
    plt.pcolor(mat, cmap=cmap)
    plt.colorbar().set_label(label=score+' score', size=fontsize)
    plt.xticks(np.arange(0.5, mat[0].size,1), types, rotation=45)
    plt.yticks(np.arange(0.5, mat[0].size, 1), types)
    plt.xlabel('Predictor State', fontsize=fontsize)
    plt.ylabel('Observed State', fontsize=fontsize)
    if title:
        plt.title(title)
    else:
        plt.title('Detection of ' + edges + ' edges')

    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)

def adjecent_grn_score(adata,
                       path,
                       score='AUPRC',
                       edges='all',
                       loc='best',
                       fontsize=10,
                       color='r',
                       errorline_color='b',
                       elinewidth=1,
                       showfig=None,
                       savefig=None,
                       format='pdf',
                       figsize=(5,3)
                       ):
    '''
    Plot the pairwise GRN similarity between consecitive cell states along a transition

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    path: `list`
        list of cell states along the transition path
    score: `str` (default: 'AUPRC')
        metric to use, choose between 'AUROC' and 'AUPRC'
    edges: `str` (default: 'all')
        which edges to consider for similarity, choose between 'all', 'positive', 'negative'
    loc: `str` (default='best')
        location of legend. Details on legend location can be found at:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
    fontsize: `int` (default: 10)
        fontsize for figure
    color: `str` (default: 'r')
        color of plot. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    errorline_color: `str` (default: 'b')
        color of deviation bars. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    elinewidth: `float` (default: 1)
        width of deviation bars
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (5,3))
        size of figure

    Returns
    -------
    None

    '''
    assert score in ['AUROC', 'AUPRC'], 'Choose between score="AUROC" and score="AUPRC"'
    assert edges in ['all', 'positive', 'negative'], 'Choose between edges="all", "positive" or "negative"'

    types = sorted(list(set(list(adata.obs['clusters']))))
    mat = adata.uns['comparison_scores'][score + '_' + edges]

    path_score, average_score, std_score = np.zeros(len(path)-1), np.zeros(len(path)-1), np.zeros(len(path)-1)
    labs = []

    for i in range(len(path)-1):
        j, k = types.index(path[i]), types.index(path[i+1])
        path_score[i] = mat[j][k]
        average_score[i] = np.mean(np.delete(mat[j], j))
        std_score[i] = np.std(np.delete(mat[j], j))
        labs.append(path[i] + ' to\n' + path[i+1])

    plt.figure(figsize=figsize)

    plt.plot(np.arange(0, path_score.size, 1), path_score, 'o--', color=color, label='Transition')
    plt.errorbar(np.arange(0, path_score.size, 1), average_score, yerr=std_score, fmt='o--', color=errorline_color, elinewidth=elinewidth, label='Average over dataset')
    plt.xticks(np.arange(0, path_score.size, 1), labs)
    plt.ylabel(score + ' score', fontsize=fontsize)
    plt.legend(loc=loc, fontsize=fontsize)

    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


'''
functions to quantify grn similarity across cell states
'''
import numpy as np
from sklearn import metrics

def edge_detection_score(m1,
                         m2
                         ):
    '''Compute AUROC/AUPRC related metrics between matrices m1 and m2 disregarding the interaction signs

    Parameters
    ----------
    m1: `~numpy.ndarray`
        observation matrix
    m2: `~numpy.ndarray`
        ground truth matrix

    Returns
    -------
    fpr: false positive rate
    tpr: true positive rate
    auroc: area under the receiver characteristic curve
    precision: precision
    recall: recall
    auprc: area under the precision recall curve

    '''
    obs = np.abs(np.ndarray.flatten(m1))
    truth = np.abs(np.ndarray.flatten(np.sign(m2)))

    fpr, tpr, thresholds = metrics.roc_curve(truth, obs)
    auroc = metrics.auc(fpr, tpr)

    precision, recall, thresholds = metrics.precision_recall_curve(truth, obs)
    auprc = metrics.auc(recall, precision)

    return fpr, tpr, auroc, precision, recall, auprc

def sign_detection_score(m1,
                         m2
                         ):
    '''Compute AUROC/AUPRC related metrics between matrices m1 and m2 when considering the interaction signs

    Parameters
    ----------
    m1: observation matrix
    m2: ground truth matrix

    Returns
    -------
    fpr: `float`
        false positive rate
    tpr: `float`
        true positive rate
    auroc: `float`
        area under the receiver characteristic curve
    precision: `float`
        precision
    recall: `float`
        recall
    auprc: `float`
        area under the precision recall curve

    '''
    obs = np.ndarray.flatten(m1)
    truth = np.ndarray.flatten(np.sign(m2))

    # compute positive and negative ground truth arrays
    grnd_pos, grnd_neg = np.zeros(truth.size), np.zeros(truth.size)
    for i in range(grnd_pos.size):
        if truth[i] == 1:
            grnd_pos[i] = 1
        elif truth[i] == -1:
            grnd_neg[i] = -1

    fpr, tpr, auroc = dict(), dict(), dict()
    fpr[0], tpr[0], thresholds = metrics.roc_curve(grnd_pos, obs)
    fpr[1], tpr[1], thresholds = metrics.roc_curve(grnd_neg, obs, pos_label=-1)
    auroc[0], auroc[1] = metrics.auc(fpr[0], tpr[0]), metrics.auc(fpr[1], tpr[1])

    recall, precision, auprc = dict(), dict(), dict()
    precision[0], recall[0], thresholds = metrics.precision_recall_curve(grnd_pos, obs)
    precision[1], recall[1], thresholds = metrics.precision_recall_curve(grnd_neg, obs, pos_label=-1)
    auprc[0], auprc[1] = metrics.auc(recall[0], precision[0]), metrics.auc(recall[1], precision[1])

    return fpr, tpr, auroc, precision, recall, auprc


def grn_comparison(adata):
    '''Compute AUROC/AUPRC scores for all pairs of state-specific gene regulatory networks
    Results are stored in adata.uns['comparison_scores']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix

    Returns
    -------
    None

    '''
    types = sorted(list(set(list(adata.obs['clusters']))))
    auroc_tot, auprc_tot = np.zeros((len(types), len(types))), np.zeros((len(types), len(types)))
    auroc_pos, auprc_pos = np.zeros((len(types), len(types))), np.zeros((len(types), len(types)))
    auroc_neg, auprc_neg = np.zeros((len(types), len(types))), np.zeros((len(types), len(types)))

    for i in range(len(types)):
        for j in range(len(types)):
            if types[i]!=types[j]:
                m1, m2 = adata.uns['average_jac'][types[i]][0], adata.uns['average_jac'][types[j]][0]
                fpr, tpr, auroc_tot[i][j], precision, recall, auprc_tot[i][j] = edge_detection_score(m1, m2)
                fpr, tpr, auroc, precision, recall, auprc = sign_detection_score(m1, m2)
                auroc_pos[i][j], auroc_neg[i][j] = auroc[0], auroc[1]
                auprc_pos[i][j], auprc_neg[i][j] = auprc[0], auprc[1]

    scores = { 'AUROC_all': auroc_tot, 'AUPRC_all': auprc_tot, 'AUROC_positive':auroc_pos, 'AUPRC_positive':auprc_pos,
               'AUROC_negative':auroc_neg, 'AUPRC_negative': auprc_neg }
    adata.uns['comparison_scores'] = scores



'''
plotting resources for sensitivity analysis
'''
import numpy as np
import matplotlib.pyplot as plt

from . import plotting_util

def regression_sens(adata,
                   font_size=10,
                   legend_font=10,
                   title_font=12,
                   legend_loc='best',
                   showfig=None,
                   savefig=None,
                   format='pdf',
                   figsize=(12,4)
                   ):
    '''Plot the summary of spliceJAC inference as a function of regression methods and parameters

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    font_size: `int` (default: 10)
        fontsize of axes and legend labels
    legend_font: `int` (default=12)
        font size of legend
    title_font: `int` (default=12)
        font size of title
    legend_loc: `str` (default='best')
        legend location
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (12,4))
        size of figure

    Returns
    -------
    None

    '''
    [alpha_ridge, alpha_lasso, method_sens] = adata.uns['method_sens']
    x, y_reg, y_ridge, y_lasso = adata.uns['sens_coeff']

    fig = plt.figure(figsize=figsize)

    ax1 = plt.subplot(131)
    ax1.set_xscale('log')
    plt.plot(x, y_reg, '-')
    ax1.set_xlim([x[0], x[-1]])
    ax1.set_xlabel('$\\epsilon$', fontsize=font_size)
    ax1.set_ylabel('Fraction of normalized coefficients > $\\epsilon$', fontsize=font_size)
    ax1.set_title('Linear regression', fontsize=title_font)

    ax2 = plt.subplot(132)
    ax2.set_xscale('log')
    for i in range(alpha_ridge.size):
        plt.plot(x, y_ridge[i], '-', label='$\\alpha_{Ridge}$=' + str(alpha_ridge[i]))
    ax2.set_xlim([x[0], x[-1]])
    ax2.set_xlabel('$\\epsilon$', fontsize=font_size)
    ax2.legend(loc=legend_loc, fontsize=legend_font)
    ax2.set_ylabel('Fraction of normalized coefficients > $\\epsilon$', fontsize=font_size)
    ax2.set_title('Ridge regression', fontsize=title_font)

    ax3 = plt.subplot(133)
    ax3.set_xscale('log')
    for i in range(alpha_lasso.size):
        plt.plot(x, y_lasso[i], '-', label='$\\alpha_{Lasso}$=' + str(alpha_lasso[i]))
    ax3.set_xlim([x[0], x[-1]])
    ax3.set_xlabel('$\\epsilon$', fontsize=font_size)
    ax3.legend(loc=legend_loc, fontsize=legend_font)
    ax3.set_ylabel('Fraction of normalized coefficients > $\\epsilon$', fontsize=font_size)
    ax3.set_title('Lasso regression', fontsize=title_font)

    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)


def sampling_sens(adata,
                  font_size=10,
                  legend_font=10,
                  legend_loc='best',
                  showfig=None,
                  savefig=None,
                  format='pdf',
                  figsize=(12,4)
                  ):
    '''Plot the summary of spliceJAC inference sensitivity to subsampling of cells in each cell state

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    font_size: `int` (default: 10)
        fontsize of axes and legend labels
    legend_font: `int` (default=10)
        font size of legend
    legend_loc: `str` (default='best')
        legend location
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (12,4))
        size of figure

    Returns
    -------
    None

    '''
    fig = plt.figure(figsize=figsize)
    ax1 = plt.subplot(131)
    ax2 = plt.subplot(132)
    ax3 = plt.subplot(133)

    sens_dict = adata.uns['sens_summary']

    for k in sens_dict.keys():
        ax1.plot(sens_dict[k]['frac'], sens_dict[k]['sign_frac'], 'o-', label=k)
        ax2.plot(sens_dict[k]['frac'], sens_dict[k]['dist'], 'o-', label=k)
        ax3.plot(sens_dict[k]['frac'], sens_dict[k]['weighted_sign'], 'o-', label=k)

    ax1.set_xlabel('Subsampling size (cell fraction)', fontsize=font_size)
    ax2.set_xlabel('Subsampling size (cell fraction)', fontsize=font_size)
    ax3.set_xlabel('Subsampling size (cell fraction)', fontsize=font_size)
    ax1.set_ylabel('Fraction of wrong signs', fontsize=font_size)
    ax2.set_ylabel('Matrix distance', fontsize=font_size)
    ax3.set_ylabel('Weighted sign distance', fontsize=font_size)
    ax1.legend(loc=legend_loc, fontsize=legend_font)
    ax2.legend(loc=legend_loc, fontsize=legend_font)
    ax3.legend(loc=legend_loc, fontsize=legend_font)

    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)


def subsample_stability(adata,
                        font_size=10,
                        dist_color='firebrick',
                        sign_color='seagreen',
                        eig_color='mediumturquoise',
                        showfig=None,
                        savefig=None,
                        format='pdf',
                        figsize=(12,4)
                        ):
    '''Summary plot of the robustness of spliceJAC inference over multiple inferences with a fraction of the total
    number of cells in each cell state

    Details on accepted colors for pyplot boxplot can be found at:
    https://matplotlib.org/stable/gallery/color/named_colors.html

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    font_size: `int` (default: 10)
        fontsize of axes and legend labels
    dist_color: `str` (default: 'firebrick')
        face color for jacobian distance boxplot
    sign_color: `str` (default: 'seagreen')
        face color for wrong sign boxplot
    eig_color: `str` (default: 'mediumturquoise')
        face color for positive eigenvalues boxplot
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (12,4))
        size of figure

    Returns
    -------
    None

    '''
    types = sorted(list(set(list(adata.obs['clusters']))))

    dev_list = [[] for i in range(len(types))]
    wrong_list = [[] for i in range(len(types))]
    pos_eig_list = [[] for i in range(len(types))]

    for i in range(len(types)):
        dev = plotting_util.jac_regr_cons(adata, types[i])
        dev_list[i] = dev
        wrong = plotting_util.count_wrong_signs(adata, types[i])
        wrong_list[i] = wrong
        pos_eig = plotting_util.count_pos_eig(adata, types[i])
        pos_eig_list[i] = pos_eig

    fig = plt.figure(figsize=figsize)

    ax1 = plt.subplot(131)
    bpt = plt.boxplot(dev_list, patch_artist=True)
    for patch in bpt['boxes']:
        patch.set_facecolor(dist_color)
    plt.xticks(np.arange(1, len(dev_list) + 1, 1), types, rotation=45)
    plt.ylabel('Jacobian distance', fontsize=font_size)

    ax2 = plt.subplot(132)
    bpt = plt.boxplot(wrong_list, patch_artist=True)
    for patch in bpt['boxes']:
        patch.set_facecolor(sign_color)
    plt.xticks(np.arange(1, len(wrong_list) + 1, 1), types, rotation=45)
    plt.ylabel('Wrong signs (%)', fontsize=font_size)

    ax3 = plt.subplot(133)
    bpt = plt.boxplot(pos_eig_list, patch_artist=True)
    for patch in bpt['boxes']:
        patch.set_facecolor(eig_color)
    plt.xticks(np.arange(1, len(pos_eig_list) + 1, 1), types, rotation=45)
    plt.ylabel('Positive eigenvalues (%)', fontsize=font_size)

    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)



'''
__init__file for plotting library
'''

from .grn_plots import visualize_network, diff_network, diff_interactions, conserved_grn, top_conserved_int, core_GRN, bif_GRN
from .jacobian_visual import visualize_jacobian, eigen_spectrum
from .sensitivity import regression_sens, sampling_sens, subsample_stability
from .sankey_plots import tg_bif_sankey
from .signaling import plot_signaling_hubs, plot_signaling_change
from .instability import plot_trans_genes, scatter_scores, compare_scvelo_scores
from .gene_variation import gene_variation, gene_var_detail, gene_var_scatter, compare_standout_genes
from .umap_scatter import umap_scatter
from .grn_comparison import adjecent_grn_score, plot_grn_comparison
from .plotting_util import plot_setup


'''
functions to plot signaling strength and changes
'''
import numpy as np
import matplotlib.pyplot as plt

def plot_signaling_hubs(adata,
                        cluster,
                        fontsize=10,
                        top_genes=5,
                        line_width=0.5,
                        show_top_genes=True,
                        criterium='weights',
                        cmap='Reds',
                        showfig=None,
                        savefig=None,
                        format='pdf',
                        figsize=(3.5,3)
                        ):
    '''Scatterplot of genes based on their signaling scores in a cell state

    Parameters of matplotlib.pyplot.scatter are explained at:
    https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster: `str`
        cell state
    fontsize: `int` (default=10)
        fontsize of figure
    top_genes: `int` (default=5)
        number of top genes to label with gene name
    line_width: `float` (default=0.5)
        line width for scatter plot
    show_top_genes: `Bool` (default=True)
        if True, annotate the top genes
    criterium: `str` (default='weights')
        criterium to rank top genes. "weights" ranks genes based on the weighted edges of the cell state GRN,
        "edges" ranks genes based on the number of edges
    cmap: `str` (default: 'Reds')
        the pyplot colormap for the scatter plot. A list of accepted color maps can be found at:
        https://matplotlib.org/stable/tutorials/colors/colormaps.html
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
        if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        format of saved figure
    figsize: `tuple` (default:(3.5,3))
        size of figure

    Returns
    -------
    None

    '''
    assert criterium=='edges' or criterium=='weights', "Choose between criterium=='edges' or criterium=='weights'"

    if criterium=='weights':
        rec = adata.uns['signaling_scores'][cluster]['incoming']
        sen = adata.uns['signaling_scores'][cluster]['outgoing']
    elif criterium=='edges':
        rec = adata.uns['signaling_scores'][cluster]['incoming_edges']
        sen = adata.uns['signaling_scores'][cluster]['outgoing_edges']
    genes = list(adata.var_names)

    # gene expression for colormap
    S = adata[adata.obs['clusters'] == cluster].layers['spliced'].toarray()
    avgS = np.mean(S, axis=0)
    cmap_lim = np.amax(avgS)

    tot = rec+sen
    sort_ind = np.argsort(tot)
    x, y, lab = rec[sort_ind[rec.size - top_genes:]], sen[sort_ind[rec.size - top_genes:]], []

    for i in sort_ind[rec.size - top_genes:]:
        lab.append(genes[i])

    fig = plt.figure(figsize=figsize)
    ax = plt.subplot(111)
    pt = plt.scatter(rec, sen, c=avgS, cmap=cmap, vmin=0, vmax=cmap_lim, edgecolors='k', linewidths=line_width)
    plt.colorbar(pt, label='Expression level')

    if show_top_genes:
        for i in range(x.size):
            ax.annotate(lab[i], (x[i], y[i]), fontsize=fontsize)

    plt.xlim([-0.1, 1.25*np.amax(rec)])
    plt.ylim([-0.1, 1.1*np.amax(sen)])
    plt.xlabel('Incoming score', fontsize=fontsize)
    plt.ylabel('Outgoing score', fontsize=fontsize)
    plt.title(cluster, fontsize=fontsize)

    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def plot_signaling_change(adata,
                          cluster1,
                          cluster2,
                          fontsize=10,
                          top_genes=10,
                          show_top_genes=True,
                          criterium='weights',
                          logscale_fc=True,
                          x_shift=0.05,
                          y_shift=0.05,
                          cmap='coolwarm',
                          line_width=0.5,
                          showfig=None,
                          savefig=None,
                          format='pdf',
                          figsize=(3.5,3)
                          ):
    '''Scatterplot of the gene signaling changes between two cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        first cell state
    cluster2: `str`
        second cell state
    fontsize: `int` (default=10)
        fontsize of figure
    top_genes: `int` (default=10)
        number of top genes to label with gene name
    show_top_genes: `Bool` (default=True)
        if True, annotate the top genes
    criterium: `str` (default='weights')
        criterium to rank top genes. "weights" ranks genes based on the weighted edges of the cell state GRN,
        "edges" ranks genes based on the number of edges
    logscale_fc: `Bool` (default=True)
        if True, rescale signaling change scores to logarithmic scale
    x_shift: `float` (default: 0.05)
        displacement on x-axis for gene annotations
    y_shift: `float` (default: 0.05)
        displacement on y-axis for gene annotations
    cmap: `str` (default: 'coolwarm')
        the pyplot colormap for the scatter plot. A list of accepted color maps can be found at:
        https://matplotlib.org/stable/tutorials/colors/colormaps.html
    line_width: `float` (default=0.5)
        line width for scatter plot
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
        if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        format of saved figure
    figsize: `tuple` (default:(3.5,3))
        size of figure

    Returns
    -------
    None

    '''
    assert criterium == 'edges' or criterium == 'weights', "Please choose between criterium=='edges' or criterium=='weights'"
    genes = list(adata.var_names)

    if criterium == 'weights':
        rec1, sen1 = adata.uns['signaling_scores'][cluster1]['incoming'], adata.uns['signaling_scores'][cluster1][
            'outgoing']
        rec2, sen2 = adata.uns['signaling_scores'][cluster2]['incoming'], adata.uns['signaling_scores'][cluster2][
            'outgoing']
    elif criterium == 'edges':
        rec1, sen1 = adata.uns['signaling_scores'][cluster1]['incoming_edges'], adata.uns['signaling_scores'][cluster1][
            'outgoing_edges']
        rec2, sen2 = adata.uns['signaling_scores'][cluster1]['incoming_edges'], adata.uns['signaling_scores'][cluster1][
            'outgoing_edges']

    delta_rec, delta_sen = rec2-rec1, sen2-sen1


    # compute gene expression fold-change between cluster 1 to cluster 2
    data1, data2 = adata[adata.obs['clusters'] == cluster1], adata[adata.obs['clusters'] == cluster2]
    S1, S2 = data1.layers['spliced'].toarray(), data2.layers['spliced'].toarray()
    avgS1, avgS2 = np.mean(S1, axis=0), np.mean(S2, axis=0)
    fc = (avgS2-avgS1)/avgS1
    if logscale_fc:
        fc = np.sign(fc)*np.log10(np.abs(fc))

    dev = delta_rec**2 + delta_sen**2
    sort_ind = np.argsort(dev)

    x, y, lab = delta_rec[sort_ind[delta_rec.size - top_genes:]], delta_sen[sort_ind[delta_rec.size - top_genes:]], []
    for i in sort_ind[delta_rec.size - top_genes:]:
        lab.append(genes[i])


    # set axes limits for best visualization
    xmin, xmax = 1.25*np.amin(delta_rec), 1.25*np.amax(delta_rec)
    if np.abs(xmin)<xmax/2:
        xmin = -xmax / 2
    if xmax<np.abs(xmin)/2:
        xmax = -xmin / 2

    ymin, ymax = 1.25 * np.amin(delta_sen), 1.25 * np.amax(delta_sen)
    if np.abs(ymin)<ymax/3:
        ymin = -ymax / 3
    if ymax<np.abs(ymin)/3:
        ymax = -ymin / 3

    fig = plt.figure(figsize=figsize)
    ax = plt.subplot(111)

    cmap_lim = np.amax(np.abs(fc))
    pt = plt.scatter(delta_rec[sort_ind], delta_sen[sort_ind], c=fc[sort_ind], s=100, cmap=cmap, vmin=-cmap_lim, vmax=cmap_lim, edgecolors='k', linewidths=line_width)
    plt.colorbar(pt, label='Expression Fold-change')

    if show_top_genes:
        for i in sort_ind[delta_rec.size - top_genes:]:
            ax.annotate(genes[i], (delta_rec[i] + x_shift, delta_sen[i] + y_shift), fontsize=fontsize)

    plt.plot([xmin, xmax], [0,0], 'k--', lw=0.5)
    plt.plot([0, 0], [ymin, ymax], 'k--', lw=0.5)
    plt.xlim([xmin, xmax])
    plt.ylim([ymin, ymax])
    plt.xlabel('$\\Delta_{Incoming}$', fontsize=fontsize)
    plt.ylabel('$\\Delta_{Outgoing}$', fontsize=fontsize)
    plt.title(cluster1+' to '+cluster2, fontsize=fontsize)

    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)


'''
methods to export spliceJAC's results
'''
import os
import numpy as np
import pandas as pd

def grn_to_csv(adata,
               cluster,
               filename
               ):
    '''Export the GRN of clusters to a csv file

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster: `str`
        cell state
    filename: `str`
        the name of export file, including path and .csv extension

    Returns
    -------
    None

    '''
    genes = list(adata.var_names)
    n = len(genes)
    grn = adata.uns['average_jac'][cluster][0][0:n, n:].copy()

    sorted = np.flip(np.argsort(np.abs(np.ndarray.flatten(grn))))

    regulator, target, interaction = [], [], []

    for i in range(sorted.size):
        k, l = np.unravel_index(sorted[i], shape=(n, n))

        if grn[k][l] == 0.:
            break

        regulator.append(genes[k])
        target.append(genes[l])
        interaction.append(grn[k][l])

    grn_dict = {'Regulator': regulator, 'Target': target, 'Interaction': interaction}
    grn_pd = pd.DataFrame.from_dict(grn_dict)

    grn_pd.to_csv(filename, index=False)

def export_grn(adata,
               cluster,
               filename=None
               ):
    '''Export the GRN of a cell state to a csv file

    If no filename is provided, a local folder results/exported_results/ will be created (if not existing) and the GRN
    file will be saved with default name "grn" + cluster + ".csv"

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster: `str`
        cell state
    filename: `str` (default: None)
        the name of export file, including path and .csv extension

    Returns
    -------
    None

    '''
    if filename==None:
        if not os.path.exists('results'):
            os.mkdir('results')
        if not os.path.exists('results/export_output'):
            os.mkdir('results/export_output')
        filename = 'results/export_output/grn' + cluster + '.csv'

    grn_to_csv(adata, cluster, filename)


def export_transition_scores(adata,
                             filename=None
                             ):
    ''' Export gene transition scores for all transitions in a csv file

    If no filename is provided, a local folder results/exported_results/ will be created (if not existing) and the GRN
    file will be saved with default name "grn" + cluster + ".csv"

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    filename: `str` (defaul: None)
        the name of export file, including path and .csv extension

    Returns
    -------
    None

    '''
    if filename==None:
        if not os.path.exists('results'):
            os.mkdir('results')
        if not os.path.exists('results/export_output'):
            os.mkdir('results/export_output')
        filename = 'results/export_output/transition_score.csv'

    tg_dict = {'gene': list(adata.var_names)}
    for t in adata.uns['transitions'].keys():
        tg_dict[t] = np.asarray( adata.uns['transitions'][t]['weights'] )

    tg_pd = pd.DataFrame.from_dict(tg_dict)
    tg_pd.to_csv(filename, index=False)


'''
plotting functions for instability scores
'''
import numpy as np
import matplotlib.pyplot as plt

def plot_trans_genes(adata,
                     cluster1,
                     cluster2,
                     top_trans_genes=10,
                     fontsize=10,
                     color='r',
                     alpha=0.5,
                     showfig=None,
                     savefig=None,
                     format='pdf',
                     figsize=(3,3)
                     ):
    '''Plot the top transition genes between two cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        starting cell state
    cluster2: `str`
        final cell state
    top_trans_genes: `int` (default: 10)
        top genes to include in the bar plot
    fontsize: `int` (default: 10)
        fontsize for labels
    color: `str` (default: 'r')
        color for bar plot. A list of named pyplot colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    alpha: `float` (default=0.5)
        shading for bar plot in [0,1]
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (3,3))
        size of figure

    Returns
    -------
    None

    '''
    weight = adata.uns['transitions'][cluster1 + '-' + cluster2]['weights']
    genes = list(adata.var_names)

    ind = np.argsort(weight)

    data, trans_genes = [], []
    for i in range(top_trans_genes):
        trans_genes.append( genes[ind[weight.size - top_trans_genes + i]] )
        data.append( weight[ind[weight.size - top_trans_genes + i]]  )

    fig = plt.figure(figsize=figsize)
    plt.barh(np.arange(1, len(data) + 1, 1), data, height=0.8, align='center', color=color, edgecolor='k', linewidth=1, alpha=alpha)
    plt.yticks(np.arange(1, len(data) + 1, 1), trans_genes, fontsize=fontsize)
    plt.xlabel('Instability score', fontsize=fontsize)
    plt.title(cluster1 + ' to ' + cluster2, fontsize=fontsize)
    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def scatter_scores(adata,
                   cluster1,
                   cluster2,
                   fontsize=10,
                   color='b',
                   showfig=None,
                   savefig=None,
                   format='pdf',
                   figsize=(3,3)
                   ):
    '''Scatter plot to compare the spliceJAC transition scores with scanpy's DEG scores of the starting cell state

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        starting cell state
    cluster2: `str`
        final cell state
    fontsize: `int` (default: 10)
        fontsize for labels
    color: `str` (default: 'b')
        color for scatter plot. A list of named pyplot colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (3,3))
        size of figure

    Returns
    -------
    None

    '''
    # get gene instability scores along transition direction
    weight = adata.uns['transitions'][cluster1 + '-' + cluster2]['weights']

    genes = list(adata.var_names)
    deg_genes = adata.uns['rank_genes_groups']['names'][cluster1]
    scores = adata.uns['rank_genes_groups']['scores'][cluster1]
    ord_weight = []
    for d in deg_genes:
        i = genes.index(d)
        ord_weight.append(weight[i])

    plt.figure(figsize=figsize)
    plt.scatter(scores, ord_weight, c=color)
    plt.xlabel('DEG score', fontsize=fontsize)
    plt.ylabel('Instability score', fontsize=fontsize)
    plt.title(cluster1 + ' to ' + cluster2, fontsize=fontsize)
    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def compare_scvelo_scores(adata,
                          annotate=True,
                          top=5,
                          color='b',
                          panel_height=3,
                          panel_length=3.5,
                          pan_per_row=4,
                          fontsize=10,
                          showfig=None,
                          savefig=None,
                          format='pdf'
                          ):
    '''Scatter plots to compare spliceJAC's transition scores with scVelo's gene likelihood scores

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    annotate: `Bool` (default=True)
        if True, annotate the genes with highest sum of scores
    top: `int` (default=5)
        number of top genes to annotate
    color: `str` (default: 'b')
        color for scatter plot. A list of named pyplot colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    panel_height: `float` (default: 3)
        height of each panel (in inches)
    panel_length: `float` (default: 3.5)
        length of each panel (in inches)
    pan_per_row: `int` (default: 4)
        number of panels per row
    fontsize: `int` (default: 10)
        fontsize for labels
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format

    Returns
    -------
    None

    '''
    types = sorted(list(set(list(adata.obs['clusters']))))
    genes = list(adata.var_names)

    # create dictionary of scvelo scores re-ordered based on gene list (which is the order followed by spliceJAC's transition scores)
    scvelo_genes = {}
    for t1 in types:
        for t2 in types:
            if t1 + '-' + t2 in adata.uns['transitions'].keys():
                gene_rank, rank = list(adata.uns['rank_dynamical_genes']['names'][t1]), adata.uns['rank_dynamical_genes']['scores'][t1]

                # create array of rank_scores for all 50 genes
                complete_rank_score = np.zeros(len(genes))
                for i in range(complete_rank_score.size):
                    if genes[i] in gene_rank:
                        complete_rank_score[i] = rank[ gene_rank.index(genes[i]) ]
                scvelo_genes[t1 + '-' + t2] = complete_rank_score


    n = len(scvelo_genes.keys())
    clusters = sorted(list(scvelo_genes.keys()))

    nrow = int(n / pan_per_row) + 1 if n % pan_per_row > 0 else int(n / pan_per_row)
    ncol = max(n % pan_per_row, pan_per_row)

    fig = plt.figure(figsize=(panel_length * ncol, panel_height * nrow))

    for i in range(n):
        # panel coordinates for plotting
        j = int(i / pan_per_row)
        k = i % pan_per_row

        ax = plt.subplot2grid((nrow, ncol), (j, k), rowspan=1, colspan=1)

        inst_score = adata.uns['transitions'][clusters[i]]['weights']

        plt.scatter( inst_score, scvelo_genes[clusters[i]], color=color )
        plt.title(clusters[i])
        plt.xlabel('Instability score (spliceJAC)', fontsize=fontsize)
        plt.ylabel('Likelihood (scVelo)', fontsize=fontsize)

        if annotate:
            tot = inst_score + scvelo_genes[clusters[i]]
            sort_ind = np.argsort(tot)
            x, y, lab = inst_score[sort_ind[inst_score.size - top:]], scvelo_genes[clusters[i]][sort_ind[scvelo_genes[clusters[i]].size - top:]], []

            for i in sort_ind[inst_score.size - top:]:
                lab.append(genes[i])

            for i in range(x.size):
                ax.annotate(lab[i], (x[i], y[i]))


    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


'''
__init__ file for the tools library
'''

from .estimate_jacobian import estimate_jacobian
from .transitions import trans_from_PAGA, transition_genes
from .grn_comparison import grn_comparison
from .regr_method_sens import regr_method_sens
from .subsampling_sens import subsampling_sens
from .grn_statistics import grn_statistics
from .export_methods import export_grn, export_transition_scores


'''
functions for Sankey diagrams
'''
import matplotlib.pyplot as plt
import plotly.graph_objects as go

from . import plotting_util

def tg_bif_sankey(adata,
                  start,
                  end,
                  gene_colormap = plt.cm.Set3.colors,
                  width=400,
                  height=600,
                  font_size=15,
                  pad=0,
                  thickness=40,
                  label_columns=True,
                  showfig=True,
                  savefig=None,
                  format='pdf'
                  ):
    '''Plot a Sankey diagram of the top transition genes involved in different cell state transitions

    More details about static export of images in python can be found at: https://plotly.com/python/static-image-export/
    More details about the arguments of plotly objects can be fount at: https://plotly.com/python/graph-objects/

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    start: `str`
        starting cell state
    end: `list`
        list of final cell states
    gene_colormap: `pyplot colormap` (default: plt.cm.Set3.colors)
        Colormap for transition genes on the left side of the Sankey diagram. To use another colormap, provide argument
        following the same syntax: plt.cm. + chosen_colormap + .colors. A list of accepted colormaps can be found at:
        https://matplotlib.org/stable/tutorials/colors/colormaps.html
    width: `int` (default=400)
        width of plotly figure
    height: `int` (default=600)
        height of plotly figure
    font_size: `int` (default: 15)
        font size of figure
    pad: `float` (default: 0)
        vertical gap between nodes of the Sankey plot
    thickness: `float` (default: 40)
        line thickness of the Sankey plot
    label_columns: `Bool` (default=True)
        if True, label the diagram columns
    showfig: `Bool` (default=True)
        if True, show the figure
    savefig: `Bool` or `None` (default=None)
        if True, save the figure using the savefig path
    format: `str` (default='pdf')
        format of saved figure

    Returns
    -------
    None

    '''
    assert 'transitions' in adata.uns.keys(), "Please run transition analysis before calling tg_bif_sankey()"

    if 'colors' not in adata.uns.keys():
        plotting_util.plot_setup(adata)

    # get TG of each cluster
    tg, cluster = [], []
    for c in end:
        deg_list, tg_list, both_list = adata.uns['transitions'][start + '-' + c]['gene_lists']
        tg = tg + tg_list + both_list
        cluster = cluster + [c for i in range(len(tg_list)+len(both_list))]

    # construct source and target arrays
    i, i_c = 0, 0
    source, label = [], []
    for t in tg:
        if t not in label:
            source.append(i)
            # target.append(cluster[i_c])
            label.append(t)
            i = i + 1
        else:
            j = label.index(t)
            source.append(j)
            # target.append(cluster[i_c])
        i_c = i_c + 1

    target = []
    i = len(set(tg))
    type = cluster[0]
    label.append(type)
    for c in cluster:
        if c!=type:
            type=c
            label.append(type)
            i = i + 1
        target.append(i)
    value = [1 for s in source]


    genes=list(set(tg))

    # define genes colors
    col_scale = list(gene_colormap)
    if len(col_scale)>len(set(source)):
        colors = list(gene_colormap)[0:len(set(source))]
    else:
        colors = []
        for i in range( len(set(source))//len(col_scale) ):
            for c in col_scale:
                colors.append(c)
        for c in col_scale[0:len(set(source))%len(col_scale)]:
            colors.append(c)

    gene_colors = []
    for i in range(len(genes)):
        rgb_scaled = tuple([255 * c for c in colors[i]])
        gene_colors.append('rgb' + str(rgb_scaled))

    # define receiving node colors and link colors
    cluster_colors = []
    for e in end:
        rgb_scaled = tuple([ 255*c for c in adata.uns['colors'][e] ])
        cluster_colors.append('rgb' + str(rgb_scaled))

    n_s = len(set(source))
    link_color = []
    for t in target:
        link_color.append( cluster_colors[t-n_s] )

    fig = go.Figure(data=[go.Sankey(
        node=dict(
            pad=pad,  # vertical gap between nodes
            thickness=thickness,  # width of nodes
            line=dict(color="black", width=0.5),
            label=label,
            color= gene_colors + cluster_colors
        ),
        link=dict(
            source= source,
            target= target,
            value=value,
            color=link_color
        ))])

    if label_columns:
        fig.add_annotation(text='TG', xref="paper", yref="paper", x=0., y=1.1, showarrow=False, align='center')
        fig.add_annotation(text='Final State', xref="paper", yref="paper", x=1.05, y=1.1, showarrow=False, align='center')
    fig.update_layout(autosize=False, width=width, height=height, font_size=font_size)

    if savefig:
        fig.write_image(savefig, format=format)
    if showfig:
        fig.show()


'''
test the inference sensitivity to cell subsampling
'''

import numpy as np
import pandas as pd

from .aux_functions import parameter_regression

def count_sign_change(v1, v2):
    '''Count the number of sign changes between two matrices

    Parameters
    ----------
    v1: `~numpy.ndarray`
        matrix 1
    v2: `~numpy.ndarray`
        matrix 2

    Returns
    -------
    sign_frac: `float`
        fraction of changed signs

    '''
    assert v1.shape == v2.shape, 'Matrices with different shapes'
    c, n = 0, v1.shape[0]
    for i in range(n):
        for j in range(n):
            if np.sign(v1[i][j])!=np.sign(v2[i][j]):
                c = c + 1
    sign_frac = float(c)/(n*n)
    return sign_frac

def mat_distance(v1,
                 v2
                 ):
    '''Compute the distance between two matrices by summing element-wise difference

    Parameters
    ----------
    v1: `~numpy.ndarray`
        matrix 1
    v2: `~numpy.ndarray`
        matrix 2

    Returns
    -------
    mat_dist: `float`
        matrix distance normalized by number of elements

    '''
    assert v1.shape==v2.shape, 'Matrices with different shapes'
    n = v1.shape[0]
    mat_dist = np.sum( np.abs(v1-v2) )/(n*n)
    return mat_dist

def count_weight_sign(v1,
                      v2
                      ):
    '''Computes the distance between two matrices by summing element-wise difference

    Parameters
    ----------
    v1: `~numpy.ndarray`
        matrix 1
    v2: `~numpy.ndarray`
        matrix 2

    Returns
    -------
    elem_dist: `~numpy.ndarray`
        matrix distance normalized by number of elements

    '''
    assert v1.shape == v2.shape, 'Matrices with different shapes'
    c, n = 0, v1.shape[0]
    for i in range(n):
        for j in range(n):
            if np.sign(v1[i][j])!=np.sign(v2[i][j]):
                c = c + np.abs(v1[i][j]-v2[i][j])
    elem_dist = float(c)/(n*n)
    return elem_dist

def test_sub_sampling(adata,
                      cluster,
                      frac,
                      nsim=10
                      ):
    '''Test the inference of gene-gene interaction matrix with subsampling for a cluster

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster: `str`
        cluster selected for inference
    frac: `float`
        fraction of cells to randomly select between [0,1]
    nsim: `int` (default: 10)
        number of independent simulations

    Returns
    -------
    sign_frac: `~numpy.ndarray`
        fraction of correct signs
    dist: `~numpy.ndarray`
        distance between reference gene-gene interaction matrix and inferred matrix using only a fraction of cells
    weight_sign: `~numpy.ndarray`
        distance based on weighted sum of correct signs

    '''
    sel_adata = adata[adata.obs['clusters'] == cluster]
    U = sel_adata.layers['unspliced'].toarray()
    S = sel_adata.layers['spliced'].toarray()
    B_ref, C, G = parameter_regression(U, S)

    n = int( U.shape[0]*frac )
    sign_frac, dist, weight_sign = np.zeros(nsim), np.zeros(nsim), np.zeros(nsim)

    for i in range(nsim):
        keep = np.random.choice(np.arange(0, U.shape[0], 1), n, replace=False)
        U_sub, S_sub = U[keep], S[keep]
        B, C, G = parameter_regression(U_sub, S_sub)
        sign_frac[i] = count_sign_change(B, B_ref)
        dist[i] = mat_distance(B, B_ref)
        weight_sign[i] = count_weight_sign(B, B_ref)
    return sign_frac, dist, weight_sign


def subsampling_sens(adata,
                  frac = np.arange(0.1, 0.91, 0.1),
                  seed=100,
                  nsim=10
                  ):
    '''Test the inference of gene-gene interaction matrix as a function of fraction of selected cells
    Results are stored in adata.uns['sens_summary']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    frac: `~numpy.ndarray` (default: numpy.arange(0.1, 0.91, 0.1))
        fraction of cells to randomly select
    seed: `int` (default=100)
        seed for random cell selection for reproducibility
    nsim: `int` (default: 10)
        number of independent simulations

    Returns
    -------
    None

    '''
    np.random.seed(seed)

    types = list(set(list(adata.obs['clusters'])))  # assumes location of cluster labels
    sens = {}

    for i in range(len(types)):
        print('Subsampling the ' + types[i] + ' cluster...')
        sign_list, dist_list, weight_list = [], [], []
        avg_sign, avg_dist, avg_weight = np.zeros(frac.size), np.zeros(frac.size), np.zeros(frac.size)
        std_sign, std_dist, std_weight = np.zeros(frac.size), np.zeros(frac.size), np.zeros(frac.size)

        for j in range(frac.size):
            sign, dist, weight_sign = test_sub_sampling(adata, types[i], frac[j], nsim=nsim)
            avg_sign[j], avg_dist[j], avg_weight[j] = np.mean(sign), np.mean(dist), np.mean(weight_sign)
            std_sign[j], std_dist[j], std_weight[j] = np.std(sign), np.std(dist), np.std(weight_sign)

            sign_list.append(sign)
            dist_list.append(dist)
            weight_list.append(weight_sign)

        df = pd.DataFrame(data={'frac': frac, 'sign_frac': avg_sign, 'dist': avg_dist, 'weighted_sign': avg_weight})
        sens[types[i]] = df
    adata.uns['sens_summary'] = sens


'''
auxiliary functions for spliceJAC inference and analysis
'''

import numpy as np
from sklearn.linear_model import Ridge, LinearRegression, Lasso

def parameter_regression(U_data,
                         S_data,
                         method='Ridge',
                         alpha=1,
                         fit_int=True
                         ):
    '''Run regression to infer spliced-unspliced interaction coefficients

    Parameters
    ----------
    U_data: `~numpy.ndarray`
        count matrix of unspliced counts
    S_data: `~numpy.ndarray`
        count matrix of spliced counts
    method: `str` (default: Ridge)
        regression method, choose between Linear, Ridge or Lasso
    alpha: `float` (default: 1)
        regularization coefficient for Ridge and Lasso
    fit_int: `Bool` (default: True)
        if True, set the fit_intercept parameter to True

    Returns
    -------
    mat: `~numpy.ndarray`
        gene-gene interaction matrix
    interc: `~numpy.ndarray`
        intercept vector
    degr: `~numpy.ndarray`
        degradation coefficient vector

    '''
    assert method == 'Ridge' or method == 'Lasso' or method=='Linear', "Please choose method='Ridge', 'Lasso' or 'Linear' "

    if method=='Linear':
        reg = LinearRegression(fit_intercept=fit_int)
    elif method == 'Ridge':
        reg = Ridge(alpha=alpha, fit_intercept=fit_int)
    elif method == 'Lasso':
        reg = Lasso(alpha=alpha, fit_intercept=fit_int)

    ncell, ngene = U_data.shape
    mat = np.zeros((ngene, ngene))
    interc = np.zeros(ngene)
    degr = np.zeros(ngene)

    for i in range(ngene):
        S_use = np.delete(S_data, i, 1)
        reg.fit(S_use, U_data[:, i])
        coeffs = reg.coef_

        mat[i][0:i] = coeffs[0:i]
        mat[i][i + 1:] = coeffs[i:]
        interc[i] = reg.intercept_

        # fit spliced degradation rate - degradation rate sin spliceJAC are computed globally, so this step is optional
        reg_g = LinearRegression(fit_intercept=False)
        reg_g.fit(S_data[:, [i]], U_data[:, i])
        degr[i] = reg_g.coef_

    return mat, interc, degr


def estimate_degr(adata,
                  first_moment=True
                  ):
    '''Estimate degradation rate coefficient vector

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    first_moment: `Bool` (default: True)
        if True, use first moment of U and S to run regression

    Returns
    -------
    degr: `~numpy.ndarray`
        degradation coefficient vector

    '''
    if first_moment:
        U_data = adata.layers['Mu']
        S_data = adata.layers['Ms']
    else:
        U_data = adata.layers['unspliced'].toarray()
        S_data = adata.layers['spliced'].toarray()

    ncell, ngene = U_data.shape
    degr = np.zeros(ngene)

    # global fit using all cells since degradation rates are global
    for i in range(ngene):
        reg_g = LinearRegression(fit_intercept=False)
        reg_g.fit(S_data[:, [i]], U_data[:, i])
        degr[i] = reg_g.coef_
    return degr


def construct_jac(mat,
                  degr,
                  b=1
                  ):
    '''Construct a Jacobian matrix given the gene-gene interactions and degradation rates

    Parameters
    ----------
    mat: `~numpy.ndarray`
        matrix of gene-gene interactions
    degr: `~numpy.ndarray`
        degradation coefficient vector
    b: `float` (default: 1)
        splicing rate constant

    Returns
    -------
    J: Jacobian matrix

    '''
    ngene = mat.shape[0]

    jac1 = np.diag(-b*np.ones(ngene))   # unspliced-unspliced part
    jac2 = np.diag(b*np.ones(ngene))    # spliced-unspliced part
    jac3 = np.diag(-degr)               # spliced-spliced part

    J1 = np.concatenate([jac1, mat], axis=1)
    J2 = np.concatenate([jac2, jac3], axis=1)
    J = np.concatenate([J1, J2])

    return J


def set_gene_axes(adata):
    '''Set up a axes name list with unspliced and spliced genes

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix

    Returns
    -------
    None

    '''
    genes = list(adata.var_names)
    axes = ['' for i in range(2*len(genes))]
    for i in range(len(genes)):
        axes[i] = genes[i] + '_U'
        axes[i + len(genes)] = genes[i]
    adata.uns['axes'] = axes


def instability_scores(adata):
    '''
    Construct an instability score for each gene in each cluster by looking at eigenvector components in the cluster
    unstable manifold. Results are saved in adata.uns['inst_scores']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix

    Returns
    -------
    None

    '''
    assert 'jacobian_lists' in adata.uns.keys(), "Run regression before computing instability scores"

    types = sorted(list(set(list(adata.obs['clusters']))))
    ngenes = 2 * len(list(adata.var_names))

    inst_scores = {}

    for k in types:
        J_list, w_list, v_list = adata.uns['jacobian_lists'][k]
        nsim = len(J_list)
        ins_score = np.zeros((nsim, ngenes))

        for j in range(nsim):
            w, v = w_list[j], v_list[j]
            score = np.zeros(w.size)
            ind = np.argwhere(np.real(w) > 0)
            for i in ind:
                score = score + np.ndarray.flatten(np.real(v[:, i]))**2
            score = score / len(ind)
            ins_score[j] = score
        inst_scores[k] = ins_score
    adata.uns['inst_scores'] = inst_scores


'''
plotting utilities
'''
import numpy as np
import matplotlib.pyplot as plt

def plot_setup(adata,
               cmap=plt.cm.Set2.colors
               ):
    '''Assign a color to each cluster from a colormap
    Results are stored in adata.uns['colors']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cmap: `pyplot colormap` (default: plt.cm.Set2.colors)
        Colormap for cell state. To use another colormap, provide argument following the same syntax:
        plt.cm. + chosen_colormap + .colors. A list of accepted colormaps can be found at:
        https://matplotlib.org/stable/tutorials/colors/colormaps.html

    Returns
    -------
    None

    '''
    clusters = list(adata.obs['clusters'])
    types = sorted(list(set(clusters)))
    colors = list(cmap)[0:len(types)]
    adata.uns['colors'] = {types[i]: colors[i] for i in range(len(types))}

def jac_regr_cons(adata,
                  cluster
                  ):
    '''Check the consistency of Jacobian regression by computing element-wise distance between Jacobians from different
    simulations for the same cell state

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster: `str`
        cell state

    Returns
    -------
    dev: a list of element-wise difference between all pairs of Jacobians

    '''
    assert 'jacobian_lists' in adata.uns.keys(), 'Run cluster stability first'

    jac_list = adata.uns['jacobian_lists'][cluster][0]
    nsim = len(jac_list)

    dev = []
    for i in range(nsim):
        for j in range( i ):
            J1, J2 = jac_list[i], jac_list[j]
            dev.append( np.linalg.norm(np.abs(J1 - J2)) )
    return dev


def compare_wrong_signs(j1,
                        j2,
                        eps=0.9
                        ):
    '''
    Count fraction of wrong signs between two matrices

    Parameters
    ----------
    j1: `~numpy.ndarray`
        matrix 1
    j2: `~numpy.ndarray`
        matrix 2
    eps: `float` (default: 0.9)
        cutoff quantile to select only the largest elements (in absolute value) in the two matrices in the range [0,1]

    Returns
    -------
    perc_sign: percentage of wrong signs between the two matrices

    '''
    n = j1.shape[0]
    j = np.sort(np.abs(np.ndarray.flatten(np.concatenate((j1, j2)))))
    t = j[int(eps*j.size)]

    count = 0
    for i in range(n):
        for j in range(n):
            if np.abs(j1[i][j] ) >t or np.abs(j2[i][j] ) >t:
                if np.sign(j1[i][j] )!=np.sign(j2[i][j]):
                    count = count + 1
    perc_sign = 100*count/float(( 1 -eps ) * n *n)
    return perc_sign


def count_wrong_signs(adata,
                      cluster,
                      eps=0.9
                      ):
    '''Quantify the fraction of wring signs between Jacobians of different simulations for the same cell state

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster: `str`
        cell state
    eps: `float` (default: 0.9)
        cutoff quantile to select only the largest elements (in absolute value) in the two matrices in the range [0,1]

    Returns
    -------
    wrong_sign: a list of wrong sign percentages between all pairs of Jacobians

    '''
    assert 'jacobian_lists' in adata.uns.keys(), 'Run cluster stability first'

    jac_list = adata.uns['jacobian_lists'][cluster][0]
    nsim = len(jac_list)

    wrong_sign = []
    for i in range(nsim):
        for j in range(i):
            J1, J2 = jac_list[i], jac_list[j]
            wrong_sign.append( compare_wrong_signs(J1, J2, eps=eps) )
    return wrong_sign


def count_pos_eig(adata,
                  cluster
                  ):
    '''
    Count the number of positive eigenvalues of the Jacobians of different simulations for the same cell state

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster: `str`
        cell state

    Returns
    -------
    pos_eig: list with number of positive eigenvalues for each Jacobian matrix

    '''
    assert 'jacobian_lists' in adata.uns.keys(), 'Run cluster stability first'

    w_list = adata.uns['jacobian_lists'][cluster][1]
    nsim = len(w_list)
    m = 2 * len(list(adata.var_names))

    pos_eig = []
    for i in range(nsim):
        w = w_list[i]
        pos_eig.append( 100 *np.real(w)[np.real(w) >0].size/float(m) )
    return pos_eig



'''
functions to identify transition driver genes
'''
import numpy as np

def find_dir(w,
             v,
             dir_method='top_eig',
             eig_number=5
             ):
    '''Identify the unstable transition directions given the eigenspectrum of the starting cell state

    Parameters
    ----------
    w: `~numpy.ndarray`
        eigenvalues of cell state Jacobian
    v: `~numpy.ndarray`
        eigenvectors of cell state Jacobian
    dir_method: `str` (default: 'top_eig')
        method to select the unstable directions, choose between 'top_eig' and 'positive'.
        'top_eig' uses the largest eigenvalues irrespective of sign; 'positive' strictly uses positive eigenvalues
    eig_number: `int` (default: 5)
        number of largest eigenvalues to consider, required for dir_method='top_eig'

    Returns
    -------
    dir: `~numpy.ndarray`
        set of unstable directions

    '''
    ind = np.flip(np.argsort(np.real(w)))
    dir = np.zeros(w.size)

    if dir_method == 'top_eig':
        for i in range(eig_number):
            dir = dir + np.real(np.ndarray.flatten(v[:, ind[i]]))
        dir = dir/eig_number

    # look only at positive eigenvalues, raise an error of no positive eigenvalues were detected
    elif dir_method == 'positive':
        w = np.real(w)
        if np.amax(w)<0.:
            raise Exception("No positive eigenvalues were detected, please set dir_method='top_eig'")
        else:
            for i in range(w.size):
                if w[i] > 0.:
                    dir = dir + np.real(np.ndarray.flatten(v[:, i]))
    return dir


def find_trans_genes(adata,
                     cluster1,
                     cluster2,
                     dir_method='top_eig',
                     eig_number=5,
                     first_moment=True
                     ):
    '''Compute the gene instability scores for transition from cluster1 to cluster2

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        starting cell state
    cluster2: `str`
        final cell states
    dir_method: `str` (default: 'top_eig')
        method to select the unstable directions, choose between 'top_eig' and 'positive'.
        'top_eig' uses the largest eigenvalues irrespective of sign; 'positive' strictly uses positive eigenvalues
    eig_number: `int` (default: 5)
        number of largest eigenvalues to consider, required for dir_method='top_eig'
    first_moment: `Bool` (default: True)
        if True, use first moments of unspliced/spliced counts

    Returns
    -------
    weight: `~numpy.ndarray`
        weight of each gene for the specified transition

    '''
    assert dir_method=='top_eig' or dir_method=='positive', "Please choose between dir_method='top_eig' or dir_method='positive'"

    if 'transitions' not in adata.uns.keys():
        adata.uns['transitions'] = {}

    # compute centers of clusters
    data1 = adata[adata.obs['clusters'] == cluster1]
    data2 = adata[adata.obs['clusters'] == cluster2]
    if first_moment:
        U1, S1 = data1.layers['Mu'], data1.layers['Ms']
        U2, S2 = data2.layers['Mu'], data2.layers['Ms']
    else:
        U1, S1 = data1.layers['unspliced'].toarray(), data1.layers['spliced'].toarray()
        U2, S2 = data2.layers['unspliced'].toarray(), data2.layers['spliced'].toarray()

    avgU1, avgS1 = np.mean(U1, axis=0), np.mean(S1, axis=0)
    avgU2, avgS2 = np.mean(U2, axis=0), np.mean(S2, axis=0)
    C1, C2 = np.concatenate((avgU1, avgS1)), np.concatenate((avgU2, avgS2))
    b = (C2-C1)/np.linalg.norm(C2-C1)

    # select eigenvalues, eigenvectors of clusters
    avg_jac = adata.uns['average_jac']
    w, v = avg_jac[cluster1][1], avg_jac[cluster1][2]
    dir = find_dir(w, v, dir_method=dir_method, eig_number=eig_number)
    proj = np.sum(dir * b) * b
    # normalize weights
    m = int(proj.size / 2)
    proj = proj[0:m] + proj[m:]
    proj = proj / np.linalg.norm(proj)
    weight = np.sqrt(proj ** 2)
    return weight


def select_top_trans_genes(adata,
                     cluster1,
                     cluster2,
                     top_DEG=5,
                     top_TG=5
                     ):
    '''Returns lists of top differentially expressed genes (DEG), transition genes (TF), and both for transition from
    cluster1 to cluster2

    If some genes are both top DEG and top TG, they are classified in the both_list,
    and n genes are selected until a total of top_DEG+top_TG genes are selected
    Results are stored in adata.uns['transitions']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        starting cell state
    cluster2: `str`
        final cell states
    top_DEG: `int` (default: 5)
        number of top DEG to select
    top_TG: `int` (default: 5)
        number of top TG to select

    Returns
    -------
    None

    '''
    assert 'rank_genes_groups' in adata.uns.keys(), "Please run 'sc.tl.rank_genes_groups()' first"

    weight = adata.uns['transitions'][cluster1 + '-' + cluster2]['weights']
    genes = list(adata.var_names)

    deg_genes = adata.uns['rank_genes_groups']['names'][cluster1]
    scores = adata.uns['rank_genes_groups']['scores'][cluster1]
    ord_weight = []
    for d in deg_genes:
        i = genes.index(d)
        ord_weight.append(weight[i])
    ord_weight = np.asarray(ord_weight)

    i_s, i_w = np.flip(np.argsort(scores)), np.flip(np.argsort(ord_weight))
    deg_list, tg_list, both_list = [], [], []

    for i in range(top_DEG):
        deg_list.append(deg_genes[i_s[i]])
    for i in range(top_TG):
        tg_list.append(deg_genes[i_w[i]])

    both_list = list(set(deg_list) & set(tg_list))
    for b in both_list:
        deg_list.remove(b)
        tg_list.remove(b)

    if len(deg_list) + len(tg_list) + len(both_list) < top_DEG + top_TG:
        cum_score = np.sqrt((ord_weight / np.amax(ord_weight))**2 + ((scores - np.amin(scores)) / (np.amax(scores) - np.amin(scores)))**2)
        cum_ind = np.flip(np.argsort(cum_score))
        for c in cum_ind:
            if deg_genes[c] not in deg_list+tg_list+both_list:
                a, b = (ord_weight[c]/np.amax(ord_weight))/cum_score[c], ((scores[c] - np.amin(scores))/(np.amax(scores) - np.amin(scores)))/cum_score[c]
                if a>b:
                    tg_list.append(deg_genes[c])
                else:
                    deg_list.append(deg_genes[c])
            if len(deg_list)+len(tg_list)+len(both_list)==(top_DEG+top_TG):
                break
    adata.uns['transitions'][cluster1 + '-' + cluster2]['gene_lists'] = [deg_list, tg_list, both_list]


def transition_genes(adata,
                     cluster1,
                     cluster2,
                     dir_method='top_eig',
                     eig_number=5,
                     top_DEG=5,
                     top_TG=5,
                     first_moment=True
                     ):
    '''Compute the gene instability scores for transition from cluster1 to cluster2
    Results are stored in adata.uns['transitions']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        starting cell state
    cluster2: `str`
        final cell states
    dir_method: `str` (default: 'top_eig')
        method to select the unstable directions, choose between 'top_eig' and 'positive'.
        'top_eig' uses the largest eigenvalues irrespective of sign; 'positive' strictly uses positive eigenvalues
    eig_number: `int` (default: 5)
        number of largest eigenvalues to consider, required for dir_method='top_eig'
    top_DEG: `int` (default: 5)
        number of top DEG to select
    top_TG: `int` (default: 5)
        number of top TG to select
    first_moment: `Bool` (default: True)
        if True, use first moments of unspliced/spliced counts

    Returns
    -------
    None

    '''
    weights = find_trans_genes(adata, cluster1, cluster2, dir_method=dir_method, eig_number=eig_number, first_moment=first_moment)
    trans_dct = {'weights': weights}
    adata.uns['transitions'][cluster1 + '-' + cluster2] = trans_dct
    select_top_trans_genes(adata, cluster1, cluster2, top_DEG=top_DEG, top_TG=top_TG)


def trans_from_PAGA(adata,
                    dir_method='top_eig',
                    eig_number=5,
                    top_DEG=5,
                    top_TG=5,
                    first_moment=True
                    ):
    '''Compute the gene instability scores for all transitions identified with PAGA
    PAGA transitions must be stored as a dataframe in adata.uns['PAGA_paths']
    results are stored in adata.uns['transitions']

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    dir_method: `str` (default: 'top_eig')
        method to select the unstable directions, choose between 'top_eig' and 'positive'.
        'top_eig' uses the largest eigenvalues irrespective of sign; 'positive' strictly uses positive eigenvalues
    eig_number: `int` (default: 5)
        number of largest eigenvalues to consider, required for dir_method='top_eig'
    top_DEG: `int` (default: 5)
        number of top DEG to select
    top_TG: `int` (default: 5)
        number of top TG to select
    first_moment: `Bool` (default: True)
        if True, use first moments of unspliced/spliced counts

    Returns
    -------
    None

    '''
    trans_df = adata.uns['PAGA_paths']
    clusters = list(trans_df.index)

    rates = trans_df.to_numpy()
    for i in range(len(clusters)):
        for j in range(len(clusters)):
            if rates[i][j]>0.:
                weights = find_trans_genes(adata, clusters[i], clusters[j], dir_method=dir_method, eig_number=eig_number, first_moment=first_moment)
                trans_dct = {'weights': weights}
                adata.uns['transitions'][clusters[i] + '-' + clusters[j]] = trans_dct
                select_top_trans_genes(adata, clusters[i], clusters[j], top_DEG=top_DEG, top_TG=top_TG)





'''
plotting resources for GRN visualization
'''
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib as mpl
import networkx as nx
from pyvis.network import Network


def plot_grn(G,
             node_size=200,
             edge_width=1,
             font_size=8,
             adata=None,
             node_color='centrality',
             cluster_name=None,
             pos_style='spring',
             base_node_size=300,
             diff_node_size=600,
             pos_edge_color='b',
             neg_edge_color='r',
             arrowsize=10,
             arrow_alpha=0.75,
             conn_style='straight',
             colorbar=True,
             fontweight='normal'
             ):
    '''Plot the GRN with positive and negative interactions

    Parameters
    ----------
    G: `networkx graph object`
        network to plot
    node_size: `int` or `str` (default: 200)
        size of nodes. If node_size='expression', the node size is scaled based on gene expression.
        Otherwise, node_size can be a fixed integer
    edge_width: `float` (default: 1)
        width of connection arrows
    fontsize: `int` (default: 10)
        fontsize for node labels.
    adata: `~anndata.AnnData` or `None` (default: `None`)
        count matrix, must be provided if node_size == 'expression'
    node_color: `str` (default: 'centrality')
        color of nodes. If node_color='centrality', the color is based on the node centrality.
        Otherwise, a matplotlib color name can be provided. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    cluster_name: `str` or `None` (default: `None`)
        name of considered cluster (must be provided if node_size='expression')
    pos_style: `str` (default: 'spring')
        position of nodes. The options are 'spring' and 'circle'. 'spring' uses the networkx spring position function,
        whereas 'circle' arranges the nodes in a circle.
    base_node_size: `float` (default: 300)
        Minimum node size (used if node_size='expression')
    diff_node_size: `float` (default: 600)
        difference between minimum and maximal node size (used if node_size='expression')
    pos_edge_color: `str` (default: 'b')
        color for positive regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    neg_edge_color: `str` (default: 'r')
        color for negative regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    arrowsize: `float` (default: 10)
        size of interaction arrows
    arrow_alpha: `float` (default: 0.75)
        shading of interaction arrows in [0,1]
    conn_style: `str` (default: 'straight')
        style of interaction arrows. The admissible styles for networkx graphs can be found at:
        https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html
    colorbar: `Bool` (default: True)
        if True, show colorbar (required if node_size='expression')
    fontweight: `str` (default: 'normal')
        style of text. Can select 'bold' for bold text.

    Returns
    -------
    None

    '''
    assert pos_style=='spring' or pos_style=='circle', "Please choose between pos_style=='spring' or pos_style=='circle'"

    if pos_style == 'spring':
        pos = nx.spring_layout(G, seed=0)
    else:
        pos = circle_pos(G)

    if conn_style=='straight':
        connectionstyle='arc3'
    elif conn_style=='curved':
        connectionstyle = 'arc3, rad=0.1'

    if node_size == 'expression':
        node_size = np.array((diff_node_size*NormalizeData(robust_mean(adata[adata.obs['clusters'].isin([cluster_name]), list(G)].X.toarray())) + base_node_size)).flatten()

    if node_color == 'centrality':
        centrality = nx.betweenness_centrality(G, k=10, endpoints=True)
        node_color = list(centrality.values())
        norm = mpl.colors.Normalize(vmin=0, vmax=1)
        cbar_label = 'Betweenness Centrality'
        cmap = plt.cm.viridis
    elif isinstance(node_color, list):
        norm = mpl.colors.Normalize(vmin=min(node_color), vmax=max(node_color))
        cbar_label = 'Expression change'
        cmap = plt.cm.BrBG

    nx.draw_networkx_nodes(G, pos, node_size=node_size, node_color=node_color, alpha=0.5, cmap=cmap)
    sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
    sm.set_array([])
    if colorbar:
        cbar = plt.colorbar(sm, label=cbar_label)

    epos = []
    eneg = []
    wpos = []
    wneg = []

    for (u, v, d) in G.edges(data=True):
        if d["weight"] > 0:
            epos.append((u, v))
            wpos.append(d["weight"])

        elif d["weight"] < 0:
            eneg.append((u, v))
            wneg.append(d["weight"])

    if edge_width == 'weight':
        edge_width_pos = NormalizeData(np.array(wpos))
        edge_width_neg = NormalizeData(-np.array(wneg))
    else:
        edge_width_pos = edge_width
        edge_width_neg = edge_width

    nx.draw_networkx_edges(G, pos, edgelist=epos, width=edge_width_pos + 0.5, edge_color=pos_edge_color,
                           arrowsize=arrowsize, alpha=arrow_alpha, connectionstyle=connectionstyle)
    nx.draw_networkx_edges(G, pos, edgelist=eneg, width=edge_width_neg + 0.5, edge_color=neg_edge_color,
                           arrowstyle="-[", arrowsize=arrowsize, alpha=arrow_alpha, connectionstyle=connectionstyle)
    nx.draw_networkx_labels(G, pos, font_size=font_size, font_family="sans-serif", font_weight=fontweight)


def visualize_network(adata,
                      cluster_name,
                      genes=None,
                      cc_id=0,
                      node_size='expression',
                      edge_width='weight',
                      font_size=10,
                      plot_interactive=True,
                      weight_quantile=.5,
                      node_color='centrality',
                      pos_style='spring',
                      title=True,
                      base_node_size=300,
                      diff_node_size=600,
                      pos_edge_color='b',
                      neg_edge_color='r',
                      arrowsize=10,
                      arrow_alpha=0.75,
                      conn_style='straight',
                      colorbar=True,
                      fontweight='normal',
                      showfig=None,
                      savefig=None,
                      format='pdf',
                      figsize=(4, 3)
                      ):
    '''
    Plot the gene regulatory network of a cluster

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster_name: `str`
        cell state
    genes: `list` (default: `None`)
        list of genes to include. If None, all genes are included
    cc_id: `int` (default: 0)
        connected component of the GRN to plot
    node_size: `int` or `str` (default: 200)
        size of nodes. If node_size='expression', the node size is scaled based on gene expression.
        Otherwise, node_size can be a fixed integer
    edge_width: `float` (default: 1)
        width of connection arrows
    fontsize: `int` (default: 10)
        fontsize for node labels.
    plot_interactive: `Bool` (default=True)
        plot the GRN interactively in notebook of figure
    weight_quantile: `float` (default=0.5)
        threshold to filter weak interactions between 0 and 1
    node_color: `str` (default: 'centrality')
        color of nodes. If node_color='centrality', the color is based on the node centrality.
        Otherwise, a matplotlib color name can be provided. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    pos_style: `str` (default: 'spring')
        position of nodes. The options are 'spring' and 'circle'. 'spring' uses the networkx spring position function,
        whereas 'circle' arranges the nodes in a circle.
    title: `Bool` (default=True)
        if True, plot title
    base_node_size: `float` (default: 300)
        Minimum node size (used if node_size='expression')
    diff_node_size: `float` (default: 600)
        difference between minimum and maximal node size (used if node_size='expression')
    pos_edge_color: `str` (default: 'b')
        color for positive regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    neg_edge_color: `str` (default: 'r')
        color for negative regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    arrowsize: `float` (default: 10)
        size of interaction arrows
    arrow_alpha: `float` (default: 0.75)
        shading of interaction arrows in [0,1]
    conn_style: `str` (default: 'straight')
        style of interaction arrows. The admissible styles for networkx graphs can be found at:
        https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html
    colorbar: `Bool` (default: True)
        if True, show colorbar (required if node_size='expression')
    fontweight: `str` (default: 'normal')
        style of text. Can select 'bold' for bold text.
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (5,4))
        size of figure

    Returns
    -------
    None

    '''
    assert conn_style=='straight' or conn_style=='curved', "Choose conn_style=='straight' or conn_style=='curved'"

    if genes == None:
        genes = list(adata.var_names)
    n = len(genes)
    A = adata.uns['average_jac'][cluster_name][0][0:n, n:].copy().T

    q_pos = np.quantile(A[A > 0], weight_quantile)
    q_neg = np.quantile(A[A < 0], 1 - weight_quantile)
    A[(A > q_neg) & (A < q_pos)] = 0

    G = nx.from_numpy_matrix(A, create_using=nx.DiGraph)
    nx.relabel_nodes(G, dict(zip(range(len(genes)), genes)), copy=False)

    G_undirected = G.to_undirected()
    subgraphs = [G.subgraph(c) for c in sorted(nx.connected_components(G_undirected), key=len, reverse=True) if
                 len(c) > 1]

    if len(subgraphs) > 1:
        print("There exist multiple connected components. Choose the parameter cc_id to show other components")

    fig, ax = plt.subplots(figsize=figsize)
    plot_grn(subgraphs[cc_id], node_size=node_size, edge_width=edge_width, font_size=font_size, adata=adata,
             node_color=node_color, cluster_name=cluster_name, pos_style=pos_style,
             base_node_size=base_node_size, diff_node_size=diff_node_size, pos_edge_color=pos_edge_color,
             neg_edge_color=neg_edge_color, arrowsize=arrowsize, arrow_alpha=arrow_alpha, conn_style=conn_style,
             colorbar=colorbar, fontweight=fontweight)

    # Title/legend
    font = {"color": "k", "fontweight": "bold", "fontsize": 20}
    if title:
        plt.title(cluster_name+' core GRN')

    # Resize figure for label readibility
    fig.tight_layout()
    plt.axis("off")
    if showfig:
        plt.show()

    if plot_interactive:
        nt = Network(notebook=True, directed=True)
        nt.from_nx(subgraphs[cc_id])
        display(nt.show('nx.html'))
    if savefig:
        plt.savefig(savefig, format=format)



def diff_network(adata,
                 cluster1,
                 cluster2,
                 genes=None,
                 cc_id=0,
                 node_size=500,
                 edge_width='weight',
                 font_size=10,
                 weight_quantile=.5,
                 pos_style='spring',
                 base_node_size=300,
                 diff_node_size=600,
                 pos_edge_color='b',
                 neg_edge_color='r',
                 arrowsize=10,
                 arrow_alpha=0.75,
                 conn_style='straight',
                 colorbar=True,
                 fontweight='normal',
                 title=True,
                 showfig=None,
                 savefig=None,
                 format='pdf',
                 figsize=(3.5, 3)
                 ):
    '''
    Plot the differential network between two cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        first cell state
    cluster2: `str`
        second cell state
    genes: `list` (default: `None`)
        list of genes to include. If None, all genes are included
    cc_id: `int` (default: 0)
        connected component of the GRN to plot
    node_size: `float` (default=500)
        size of nodes in the GRN
    edge_width: `float` or `str` (default='weight')
         width of GRN edges. If edge_width='weight', the edge width is proportional to the interaction strength.
    fontsize: `int` (default: 10)
        fontsize for figure.
    weight_quantile: `float` (default: 0.5)
        threshold to filter weak interactions between 0 and 1
    pos_style: `str` (default: 'spring')
        position of nodes. The options are 'spring' and 'circle'. 'spring' uses the networkx spring position function,
        whereas 'circle' arranges the nodes in a circle.
    base_node_size: `float` (default: 300)
        Minimum node size (used if node_size='expression')
    diff_node_size: `float` (default: 600)
        difference between minimum and maximal node size (used if node_size='expression')
    pos_edge_color: `str` (default: 'b')
        color for positive regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    neg_edge_color: `str` (default: 'r')
        color for negative regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    arrowsize: `float` (default: 10)
        size of interaction arrows
    arrow_alpha: `float` (default: 0.75)
        shading of interaction arrows in [0,1]
    conn_style: `str` (default: 'straight')
        style of interaction arrows. The admissible styles for networkx graphs can be found at:
        https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html
    colorbar: `Bool` (default: True)
        if True, show colorbar (required if node_size='expression')
    fontweight: `str` (default: 'normal')
        style of text. Can select 'bold' for bold text.
    title: `Bool` (default: True)
        if True, plot title
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (3.5,3))
        size of figure

    Returns
    -------
    None

    '''
    assert conn_style == 'straight' or conn_style == 'curved', \
        "Please choose between conn_style=='straight' and conn_style=='curved'"

    if genes == None:
        genes = list(adata.var_names)
    n = len(genes)

    A1 = adata.uns['average_jac'][cluster1][0][0:n, n:].copy().T
    A2 = adata.uns['average_jac'][cluster2][0][0:n, n:].copy().T
    A = A2 - A1

    q_pos = np.quantile(A[A > 0], weight_quantile)
    q_neg = np.quantile(A[A < 0], 1 - weight_quantile)
    A[(A > q_neg) & (A < q_pos)] = 0

    G = nx.from_numpy_matrix(A, create_using=nx.DiGraph)
    nx.relabel_nodes(G, dict(zip(range(len(genes)), genes)), copy=False)

    G_undirected = G.to_undirected()
    subgraphs = [G.subgraph(c) for c in sorted(nx.connected_components(G_undirected), key=len, reverse=True) if
                 len(c) > 1]

    if len(subgraphs) > 1:
        print("There exist multiple connected components. Choose the parameter cc_id to show other components")

    fig, ax = plt.subplots(figsize=figsize)

    data1 = adata[adata.obs['clusters'] == cluster1, list(subgraphs[cc_id])].X.toarray()
    data2 = adata[adata.obs['clusters'] == cluster2, list(subgraphs[cc_id])].X.toarray()
    expr_change = list(np.mean(data2, axis=0) - np.mean(data1, axis=0))

    plot_grn(subgraphs[cc_id], node_size=node_size, edge_width=edge_width, font_size=font_size, adata=adata,
             node_color=expr_change, cluster_name=None, pos_style=pos_style, base_node_size=base_node_size,
             diff_node_size=diff_node_size, pos_edge_color=pos_edge_color, neg_edge_color=neg_edge_color,
             arrowsize=arrowsize, arrow_alpha=arrow_alpha, conn_style=conn_style, colorbar=colorbar,
             fontweight=fontweight)
    plt.axis("off")

    # Title/legend
    font = {"color": "k", "fontweight": "bold", "fontsize": 20}
    if title:
        plt.title(cluster1 + '-' + cluster2 + ' differential GRN')

    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def diff_interactions(adata,
                      cluster1,
                      cluster2,
                      top_int=10,
                      loc='best',
                      title=False,
                      fontsize=10,
                      legend_font=10,
                      legend_col=1,
                      showfig=None,
                      savefig=None,
                      format='pdf',
                      figsize=(4,5)
                      ):
    '''
    Plot the top differential interactions between two cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        first cell state
    cluster2: `str`
        second cell state
    top_int: `int` (default=10)
        number of top changed interactions to plot
    loc: `str` (default='best')
        location of legend. Details on legend location can be found at:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
    title: `Bool` (default: False)
        if True, plot title
    fontsize: `int` (default: 10)
        fontsize for figure.
    legend_font: `int` (default: 10)
        fontsize of legend
    legend_col: `int` (default: 1)
        number of columns in legend
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (4,5))
        size of figure

    Returns
    -------
    None

    '''
    genes = list(adata.var_names)
    n = len(genes)

    A1 = adata.uns['average_jac'][cluster1][0][0:n, n:].copy().T
    A2 = adata.uns['average_jac'][cluster2][0][0:n, n:].copy().T
    A = A2 - A1

    y = np.ndarray.flatten(A)
    sorted = np.flip(np.argsort( np.abs(y) ))

    int_list, name = [], []
    old, new = [], []
    color = []
    for i in range(top_int):
        k, l = np.unravel_index(sorted[i], shape=(n,n))
        int_list.append(A[k][l])
        name.append( genes[k] + ' to ' + genes[l] )
        old.append(A1[k][l])
        new.append(A2[k][l])

    fig = plt.figure(figsize=figsize)

    plt.plot( old, np.arange(1, len(int_list)+1, 1), 'bo', label=cluster1 )
    plt.plot( new, np.arange(1, len(int_list) + 1, 1), 'ro', label=cluster2)
    plt.plot( np.zeros(len(int_list)), np.linspace(0, len(int_list) + 1, len(int_list)), 'k--' )

    for i in range(len(old)):
        plt.arrow( old[i], i+1, 0.95*(new[i]-old[i]), 0, head_length=0.05, head_width=0.5,
                   length_includes_head=True, color='grey')

    plt.yticks( np.arange(1, len(int_list)+1, 1), name )
    plt.xlabel('Interaction strenght change', fontsize=fontsize)
    plt.legend(loc=loc, fontsize=legend_font, ncol=legend_col)
    plt.ylim([0, len(int_list) + 2])

    if title:
        plt.title(title, fontsize=fontsize)
    else:
        plt.title(cluster1 + '-' + cluster2 + ' differential interactions', fontsize=fontsize)
    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)


def conserved_grn(adata,
                  cluster1,
                  cluster2,
                  genes=None,
                  cc_id=0,
                  node_size=500,
                  edge_width='weight',
                  font_size=10,
                  weight_quantile=.5,
                  pos_style='spring',
                  title=True,
                  base_node_size=300,
                  diff_node_size=600,
                  pos_edge_color='b',
                  neg_edge_color='r',
                  arrowsize=10,
                  arrow_alpha=0.75,
                  conn_style='straight',
                  colorbar=True,
                  fontweight='normal',
                  showfig=None,
                  savefig=None,
                  format='pdf',
                  figsize=(3.5, 3)
                  ):
    '''
    Plot the conserved network between two cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        first cell state
    cluster2: `str`
        second cell state
    genes: `list` or `None` (default: `None`)
        list of genes to consider. If None, all genes are considered
    cc_id: `int` (default: 0)
        connected component of the GRN to plot
    node_size: `int` (default: 500)
        size of nodes in the GRN
    edge_width: `str` or `int` (default: 'weight')
        width of GRN edges. If edge_width='weight', the edge width is proportional to the interaction strength.
    fontsize: `int` (default: 10)
        fontsize for figure.
    weight_quantile: `float` (default: 0.5)
        threshold to filter weak interactions between 0 and 1
    pos_style: `str` (default: 'spring')
        position of nodes. The options are 'spring' and 'circle'. 'spring' uses the networkx spring position function,
        whereas 'circle' arranges the nodes in a circle.
    title: `Bool` (default: True)
        if True, plot title
    base_node_size: `float` (default: 300)
        Minimum node size (used if node_size='expression')
    diff_node_size: `float` (default: 600)
        difference between minimum and maximal node size (used if node_size='expression')
    pos_edge_color: `str` (default: 'b')
        color for positive regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    neg_edge_color: `str` (default: 'r')
        color for negative regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    arrowsize: `float` (default: 10)
        size of interaction arrows
    arrow_alpha: `float` (default: 0.75)
        shading of interaction arrows in [0,1]
    conn_style: `str` (default: 'straight')
        style of interaction arrows. The admissible styles for networkx graphs can be found at:
        https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html
    colorbar: `Bool` (default: True)
        if True, show colorbar (required if node_size='expression')
    fontweight: `str` (default: 'normal')
        style of text. Can select 'bold' for bold text.
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (3.5,3))
        size of figure

    Returns
    -------
    None

    '''
    if genes == None:
        genes = list(adata.var_names)
    n = len(genes)

    A1 = adata.uns['average_jac'][cluster1][0][0:n, n:].copy().T
    A2 = adata.uns['average_jac'][cluster2][0][0:n, n:].copy().T

    A = np.zeros((n,n))
    for i in range(n):
        for j in range(n):
            if np.sign(A2[i][j]) != np.sign(A1[i][j]):
                A[i][j] = 0.
            else:
                A[i][j] = np.mean([A2[i][j], A1[i][j]])

    q_pos = np.quantile(A[A > 0], weight_quantile)
    q_neg = np.quantile(A[A < 0], 1 - weight_quantile)
    A[(A > q_neg) & (A < q_pos)] = 0

    G = nx.from_numpy_matrix(A, create_using=nx.DiGraph)
    nx.relabel_nodes(G, dict(zip(range(len(genes)), genes)), copy=False)

    G_undirected = G.to_undirected()
    subgraphs = [G.subgraph(c) for c in sorted(nx.connected_components(G_undirected), key=len, reverse=True) if
                 len(c) > 1]

    if len(subgraphs) > 1:
        print("There exist multiple connected components. Choose the parameter cc_id to show other components")

    fig, ax = plt.subplots(figsize=figsize)

    data1 = adata[adata.obs['clusters'] == cluster1, list(subgraphs[cc_id])].X.toarray()
    data2 = adata[adata.obs['clusters'] == cluster2, list(subgraphs[cc_id])].X.toarray()
    expr_change = list(np.mean(data2, axis=0) - np.mean(data1, axis=0))

    plot_grn(subgraphs[cc_id], node_size=node_size, edge_width=edge_width, font_size=font_size, adata=adata,
             node_color=expr_change, cluster_name=None, pos_style=pos_style, base_node_size=base_node_size,
             diff_node_size=diff_node_size,
             pos_edge_color=pos_edge_color, neg_edge_color=neg_edge_color, arrowsize=arrowsize, arrow_alpha=arrow_alpha,
             conn_style=conn_style, colorbar=colorbar, fontweight=fontweight)
    plt.axis("off")

    # Title/legend
    font = {"color": "k", "fontweight": "bold", "fontsize": 20}
    if title:
        plt.title(cluster1 + '-' + cluster2 + ' conserved GRN')

    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)


def top_conserved_int(adata,
                      cluster1,
                      cluster2,
                      top_int=10,
                      title=False,
                      fontsize=10,
                      alpha=0.5,
                      showfig=None,
                      savefig=None,
                      format='pdf',
                      figsize=(4,5)
                      ):
    '''
    Plot the top conserved interactions between two cell states

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        first cell state
    cluster2: `str`
        second cell state
    top_int: `int` (default=10)
        number of top chnaged interactions to plot
    title: `Bool` (default: False)
        if True, plot title
    fontsize: `int` (default: 10)
        fontsize for figure.
    alpha: `float` (default=0.5)
        shading of bar plot
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (4,5))
        size of figure

    Returns
    -------
    None

    '''
    genes = list(adata.var_names)
    n = len(genes)

    A1 = adata.uns['average_jac'][cluster1][0][0:n, n:].copy().T
    A2 = adata.uns['average_jac'][cluster2][0][0:n, n:].copy().T

    A = np.zeros((n, n))
    for i in range(n):
        for j in range(n):
            if np.sign(A2[i][j]) != np.sign(A1[i][j]):
                A[i][j] = 0.
            else:
                A[i][j] = np.mean([A2[i][j], A1[i][j]])


    y = np.ndarray.flatten(A)
    sorted = np.flip(np.argsort(np.abs(y)))

    int_list, name = [], []
    color = []
    for i in range(top_int):
        k, l = np.unravel_index(sorted[i], shape=(n, n))
        int_list.append(A[k][l])
        name.append(genes[k] + ' to ' + genes[l])
        color.append('b') if A[k][l]>0 else color.append('r')


    plt.figure(figsize=figsize)

    plt.barh( np.arange(len(int_list), 0, -1), int_list, color=color, alpha=alpha )

    plt.yticks( np.arange(len(int_list), 0, -1), name)
    plt.xlabel('Interaction strenght', fontsize=fontsize)
    plt.ylim([0, len(int_list) + 1])

    if title:
        plt.title(title, fontsize=fontsize)
    else:
        plt.title(cluster1 + '-' + cluster2 + ' conserved interactions', fontsize=fontsize)
    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)


def robust_mean(X):
    '''Calculate gene expression mean based on quantile selection

    Parameters
    ----------
    X: ~numpy.ndarray`
        array of gene expression values

    Returns
    -------
    robust mean: ~numpy.ndarray`
        array of mean gene expression

    '''
    rob_mean = 0.25 * np.quantile(X, q=0.75, axis=0) + 0.25 * np.quantile(X, q=0.25, axis=0) + \
               0.5 * np.quantile(X, q=0.5, axis=0)
    return rob_mean


def NormalizeData(data):
    '''normalize the gene expression data

    Parameters
    ----------
    data: ~numpy.ndarray`
        data array

    Returns
    -------
    scaled: ~numpy.ndarray`
        normalized data

    '''
    scaled = (data - np.min(data)) / (np.max(data) - np.min(data))
    return scaled


def circle_pos(G):
    '''Compute coordinates along a unit circle for GRN node positions

    Parameters
    ----------
    G: `networkx object`
        the networkx network

    Returns
    -------
    pos: ~numpy.ndarray`
        array of positions

    '''
    nodes = list(G.nodes)
    center, radius = [0,0], 1

    pos = {}
    angle = np.linspace(0, 2*np.pi, len(nodes), endpoint=False)

    for i in range(angle.size):
        coord = np.array([ center[0]+np.sin(angle[i]*radius), center[1]+np.cos(angle[i]*radius) ])
        pos[nodes[i]]=coord
    return pos


def subset_jacobian(J,
                    genes,
                    genelist
                    ):
    '''select gene-gene interactions for genes in the genelist

    Parameters
    ----------
    J: `~numpy.ndarray`
        gene-gene interaction matrix
    genes: `list`
        full list of genes
    genelist: `list`
        list of selected genes

    Returns
    -------
    B_subset: `~numpy.ndarray`
        reduced gene-gene interaction matrix

    '''
    ind = []
    for s in genelist:
        ind.append(genes.index(s))

    m = int(J.shape[0]/2)
    B = np.transpose(J[0:m, m:])
    B_subset = B[ind][:, ind]
    return B_subset

def core_GRN(adata,
             cluster1,
             cluster2,
             type_color=['orange', 'plum', 'yellowgreen'],
             pos_edge_color='b',
             neg_edge_color='r',
             node_size=500,
             node_alpha=0.5,
             arrowsize=10,
             arrow_alpha=0.75,
             conn_style='arc3, rad=0.1',
             node_font=8,
             legend=True,
             legend_font=10,
             legend_ncol=2,
             legend_loc='lower center',
             axis=False,
             xlim=[-1.2, 1.2],
             ylim=None,
             showfig=None,
             savefig=None,
             format='pdf',
             figsize=(3,3)
             ):
    ''' Plot a reduced GRN including the top DEG of the starting cluster and the top transition genes

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    cluster1: `str`
        first cell state
    cluster2: `str`
        second cell state
    type_color: `list` (default=['orange', 'plum', 'yellowgreen'])
        colors for nodes that are DEG, transition genes and both. A list of accepted colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    pos_edge_color: `str` (default: 'b')
        color for positive regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    neg_edge_color: `str` (default: 'r')
        color for negative regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    node_size: 'int' (default: 500)
        size of nodes
    node_alpha: `float` (default: 0.5)
        shading of nodes
    arrowsize: `float` (default: 10)
        size of interaction arrows
    arrow_alpha: `float` (default: 0.75)
        shading of interaction arrows in [0,1]
    conn_style: `str` (default: 'arc3, rad=0.1')
        style of interaction arrows. The admissible styles for networkx graphs can be found at:
        https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html
    node_font: `int` (default: 8)
        fontsize for node labels.
    legend: `Bool` (default=True)
        if True, include legend
    legend_font: `int` (default=10)
        font size for legend
    legend_ncol: `int` (default=2)
        number of columns in legend
    legend_loc: `str` (default: 'lower center')
        legend location.  Details on legend location can be found at:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
    axis: `Bool` (default=False)
        if True, plot axes
    xlim: `list` (default=[-1.2, 1.2])
        inteval on x-axis
    ylim: `list` or `None` (default=None)
        inteval on y-axis
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (3,3))
        size of figure

    Returns
    -------
    None

    '''
    deg_list, tg_list, both_list = adata.uns['transitions'][cluster1 + '-' + cluster2]['gene_lists']

    # 1) select reduced Jacobian with top DEG and TG
    sel_genes = deg_list+tg_list+both_list
    node_color = [type_color[0] for d in deg_list] + [type_color[1] for t in tg_list] + \
                 [type_color[2] for b in both_list]

    B_core = subset_jacobian(adata.uns['average_jac'][cluster1][0], list(adata.var_names), sel_genes)

    # 2) filter number of edges
    A = B_core.copy()
    q_pos = np.quantile(A[A > 0], 0.5)
    q_neg = np.quantile(A[A < 0], 1 - 0.5)
    A[(A > q_neg) & (A < q_pos)] = 0

    # 3) create graph from Jacobian matrix
    G = nx.from_numpy_matrix(A, create_using=nx.DiGraph)
    nx.relabel_nodes(G, dict(zip(range(len(sel_genes)), sel_genes)), copy=False)
    pos = circle_pos(G)

    # 4) select positive and negative interactions for separate arrow plotting
    epos = []
    eneg = []
    wpos = []
    wneg = []

    for (u, v, d) in G.edges(data=True):
        if d["weight"] > 0:
            epos.append((u, v))
            wpos.append(d["weight"])

        elif d["weight"] < 0:
            eneg.append((u, v))
            wneg.append(d["weight"])

    # 5) plotting
    fig = plt.figure(figsize=figsize)
    ax = plt.subplot(111)

    nx.draw_networkx_nodes(G, pos, node_size=node_size, node_color=node_color, alpha=node_alpha)
    nx.draw_networkx_edges(G, pos, edgelist=epos, edge_color=pos_edge_color, arrowsize=arrowsize,
                           alpha=arrow_alpha, connectionstyle=conn_style)
    nx.draw_networkx_edges(G, pos, edgelist=eneg, edge_color=neg_edge_color, arrowstyle="-[", arrowsize=arrowsize,
                           alpha=arrow_alpha, connectionstyle=conn_style)

    nx.draw_networkx_labels(G, pos, font_size=node_font)

    plt.xlim(xlim)
    if ylim==None:
        ylim = [-1.6, 1.2] if legend else [-1.2, 1.2]
    plt.ylim(ylim)

    if legend:
        labels = [cluster1 + ' DEG', cluster1 + ' to ' + cluster2 + ' TG','DEG+TG']
        patches = [mpatches.Patch(color=type_color[i], label=labels[i]) for i in range(len(labels))]
        plt.legend(bbox_to_anchor=(0.5, -0.1), handles=patches, loc=legend_loc, ncol=legend_ncol, fontsize=legend_font)

    if not(axis):
        plt.axis("off")
        
    plt.tight_layout()

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)




def bif_GRN(adata, start, end, pos_edge_color='b',
             neg_edge_color='r',
             node_size=750,
             node_alpha=0.5,
             arrowsize=10,
             arrow_alpha=0.75,
             arrowstyle='arc3, rad=0.1',
             node_font=10,
             legend=True,
             legend_font=10,
             legend_ncol=3,
             legend_loc='lower center',
             axis=True,
             xlim=[-1.2, 1.2],
             ylim=None,
             showfig=None,
             savefig=None,
             format='pdf',
             figsize=(6,6)
            ):
    '''Plot the reduced GRN of transition genes involved in different cell state transitions

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    start: `str`
        starting cell state
    end: `list`
        list of ending cell states
    pos_edge_color: `str` (default: 'b')
        color for positive regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    neg_edge_color: `str` (default: 'r')
        color for negative regulation arrow. A full list of accepted named colors can be found at:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    node_size: `int` (default: 500)
        size of nodes
    node_alpha: `float` (default: 0.5)
        shading of nodes
    arrowsize: `float` (default: 10)
        size of interaction arrows
    arrow_alpha: `float` (default: 0.75)
        shading of interaction arrows in [0,1]
    conn_style: `str` (default: 'arc3, rad=0.1')
        style of interaction arrows. The admissible styles for networkx graphs can be found at:
        https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html
    node_font: `int` (default=8)
        font size of node labels
    legend: `Bool` (default=True)
        if True, include legend
    legend_font: `int` (default=10)
        font size for legend
    legend_ncol: `int` (default=2)
        number of columns in legend
    legend_loc: `str` (default='lower center')
        legend location.  Details on legend location can be found at:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
    axis: `Bool` (default=False)
        if True, plot axes
    xlim: `list` (default=[-1.2, 1.2])
        inteval on x-axis
    ylim: `list` or `None` (default=None)
        inteval on y-axis
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format
    figsize: `tuple` (default: (6,6))
        size of figure

    Returns
    -------
    None

    '''
    types = sorted(list(set(list(adata.obs['clusters']))))
    colors = list(plt.cm.Set3.colors)[0:len(types)]

    # get TG of each cluster
    tg, cluster = [], []
    for c in end:
        deg_list, tg_list, both_list = adata.uns['transitions'][start + '-' + c]['gene_lists']
        tg = tg + tg_list + both_list
        cluster = cluster + [c for i in range(len(tg_list) + len(both_list))]

    tg_sel, cluster_sel, node_color = [], [], []
    for i in range(len(tg)):
        if tg[i] not in tg_sel:
            tg_sel.append(tg[i])
            cluster_sel.append(cluster[i])
            node_color.append( colors[types.index(cluster[i])] )
        else:
            j = tg_sel.index(tg[i])
            if cluster_sel[j]!='shared':
                cluster_sel[j] = 'shared'
                node_color[j] = 'lightgray'

    B_core = subset_jacobian(adata.uns['average_jac'][start][0], list(adata.var_names), tg_sel)

    # 2) filter number of edges
    A = B_core.copy()
    q_pos = np.quantile(A[A > 0], 0.5)
    q_neg = np.quantile(A[A < 0], 1 - 0.5)
    A[(A > q_neg) & (A < q_pos)] = 0

    # 3) create graph from Jacobian matrix
    G = nx.from_numpy_matrix(A, create_using=nx.DiGraph)
    nx.relabel_nodes(G, dict(zip(range(len(tg_sel)), tg_sel)), copy=False)
    pos = circle_pos(G)

    # 4) select positive and negative interactions for separate arrow plotting
    epos = []
    eneg = []
    wpos = []
    wneg = []

    for (u, v, d) in G.edges(data=True):
        if d["weight"] > 0:
            epos.append((u, v))
            wpos.append(d["weight"])

        elif d["weight"] < 0:
            eneg.append((u, v))
            wneg.append(d["weight"])

    fig = plt.figure(figsize=figsize)
    ax = plt.subplot(111)

    nx.draw_networkx_nodes(G, pos, node_size=node_size, node_color=node_color, alpha=node_alpha)
    nx.draw_networkx_edges(G, pos, edgelist=epos, edge_color=pos_edge_color, arrowsize=arrowsize,
                           alpha=arrow_alpha, connectionstyle=arrowstyle)
    nx.draw_networkx_edges(G, pos, edgelist=eneg, edge_color=neg_edge_color, arrowstyle="-[", arrowsize=arrowsize,
                           alpha=arrow_alpha, connectionstyle=arrowstyle)

    nx.draw_networkx_labels(G, pos, font_size=node_font)

    plt.xlim(xlim)
    if ylim == None:
        ylim = [-1.6, 1.2] if legend else [-1.2, 1.2]
    plt.ylim(ylim)

    if legend:
        labels = [c+' TG' for c in end] + ['Shared TG']
        type_color = []
        for c in end:
            j = types.index(c)
            type_color.append(colors[j])
        type_color.append('lightgray')

        patches = [mpatches.Patch(color=type_color[i], label=labels[i]) for i in range(len(labels))]
        plt.legend(handles=patches, loc=legend_loc, ncol=legend_ncol, fontsize=legend_font)

    if not(axis):
        plt.axis("off")

    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format, dpi=300)






'''
plotting resources for Jacobian and eigenvalues visualization
'''
import numpy as np
import matplotlib.pyplot as plt

def visualize_jacobian(adata,
                       panel_height=3,
                       panel_length=3.5,
                       pan_per_row=4,
                       fontsize=10,
                       cmap='RdBu_r',
                       showfig=None,
                       savefig=None,
                       format='pdf'
                       ):
    '''Plot the inferred gene-gene interaction matrices of each cell state

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    panel_height: `float` (default: 3)
        height of each panel (in inches)
    panel_length: `float` (default: 3.5)
        length of each panel (in inches)
    pan_per_row: `int` (default: 4)
        number of panels per row
    fontsize: `int` (default: 10)
        fontsize for labels
    cmap: `str` (default: 'RdBu_r')
        colormap for Jacobian visualization. Accepted colormaps ca be found at:
        https://matplotlib.org/stable/tutorials/colors/colormaps.html
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format

    Returns
    -------
    None

    '''
    inf_dict = adata.uns['average_jac']
    n = len(inf_dict.keys())
    clusters = sorted(list(inf_dict.keys()))

    nrow = int(n / pan_per_row) + 1 if n % pan_per_row > 0 else int(n / pan_per_row)
    ncol = max(n % pan_per_row, pan_per_row)

    fig = plt.figure(figsize=(panel_length * ncol, panel_height * nrow))

    for i in range(n):
        # panel coordinates for plotting
        j = int(i / pan_per_row)
        k = i % pan_per_row

        ax = plt.subplot2grid((nrow, ncol), (j, k), rowspan=1, colspan=1)

        mat = inf_dict[clusters[i]][0]
        m = int(mat.shape[0] / 2)
        J = mat[0:m, m:]
        lim = np.amax(np.abs(J))
        x = np.arange(0, J.shape[0] + 1, 1)
        pt = ax.pcolor(x, x, J, cmap=cmap, vmin=-lim, vmax=lim)
        cbar = plt.colorbar(pt, label='coefficient')
        plt.title(clusters[i])
        plt.xlabel('Regulator gene', fontsize=fontsize)
        plt.ylabel('Target gene', fontsize=fontsize)

        # matrix_heatmap(ax, inf_dict[clusters[i]][0], title=clusters[i])

    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)


def eigen_spectrum(adata,
                   panel_height=4,
                   panel_length=4,
                   pan_per_row=4,
                   fontsize=10,
                   show_frac=True,
                   loc='lower right',
                   show_zero=True,
                   showfig=None,
                   savefig=None,
                   format='pdf'
                   ):
    '''Plot the eigenvalues of each cell state

    Parameters
    ----------
    adata: `~anndata.AnnData`
        count matrix
    panel_height: `float` (default: 3)
        height of each panel (in inches)
    panel_length: `float` (default: 3.5)
        length of each panel (in inches)
    pan_per_row: `int` (default: 4)
        number of panels per row
    fontsize: `int` (default: 10)
        fontsize for labels
    show_frac: `Bool` (default: True)
        show legend with fraction of positive eigenvalues
    loc: `str` (default: 'lower right')
        location of legend. Accepted legend positions can be found at:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
    show_zero: `Bool` (default: True)
        plot horizontal line to highlight change of sign
    showfig: `Bool` or `None` (default: `None`)
        if True, show the figure
    savefig: `Bool` or `None` (default: `None`)
         if True, save the figure using the savefig path
    format: `str` (default: 'pdf')
        figure format

    Returns
    -------
    None

    '''
    inf_dict = adata.uns['average_jac']
    n = len(inf_dict.keys())
    clusters = sorted(list(inf_dict.keys()))

    nrow = int(n / pan_per_row) + 1 if n % pan_per_row > 0 else int(n / pan_per_row)
    ncol = max(n % pan_per_row, pan_per_row)

    fig = plt.figure(figsize=(panel_length*ncol, panel_height*nrow))

    for i in range(n):

        # panel coordinates for plotting
        j = int(i / pan_per_row)
        k = i % pan_per_row

        ax = plt.subplot2grid((nrow, ncol), (j, k), rowspan=1, colspan=1)
        spectrum_plot(ax, np.sort(np.real(inf_dict[clusters[i]][1])), clusters[i], show_frac=show_frac, loc=loc, show_zero=show_zero, fontsize=fontsize)

    plt.tight_layout()
    if showfig:
        plt.show()
    if savefig:
        plt.savefig(savefig, format=format)


def spectrum_plot(ax,
                  v,
                  title,
                  show_frac=True,
                  loc='lower right',
                  show_zero=True,
                  fontsize=10
                  ):
    '''Plot the eigenspectrum of a cell state on a matplotlib axis

    Parameters
    ----------
    ax: pyplot axis
        axis to plot
    v: `~numpy.ndarray`
        ranked real components of eigenvalues
    title: `str`
        title of plot
    show_frac: `Bool` (default: True)
        show legend with fraction of positive eigenvalues
    loc: `str` (default: 'lower right')
        location of legend. Accepted legend positions can be found at:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
    show_zero: `Bool` (default: True)
        plot horizontal line to highlight change of sign
    fontsize: `int` (default: 10)
        fontsize for labels

    Returns
    -------
    None

    '''
    if show_frac:
        frac = round(100*float(v[v>0].size)/v.size, 2)
        ax.plot(np.arange(1, v.size + 1, 1), v, 'ko', label='$\\rho=$'+str(frac)+'%')
        plt.legend(loc=loc)
    else:
        ax.plot(np.arange(1, v.size + 1, 1), v, 'ko')
    if show_zero:
        ax.plot([1, v.size], [0, 0], 'r--')
    plt.xlabel('Eigenvalue rank', fontsize=fontsize)
    plt.ylabel('Eigenvalue', fontsize=fontsize)
    plt.xlim([1, v.size])
    if title:
        plt.title(title, fontsize=fontsize)